from dataclasses import dataclass
from typing import List, Union

import numpy as np
import torch

from ....preprocessors import AudioFeatureExtractor, AudioFeatureExtractorConfig
from ....registry import register_preprocessor
from ....utils import (
    Logger,
    convert_batch_dict_dtype,
    mel_filter_bank,
    spectrogram,
    window_function,
)


logger = Logger(__name__)


@dataclass
class WhisperFeatureExtractorConfig(AudioFeatureExtractorConfig):
    name = "whisper_feature_extractor"
    feature_size: int = 80
    sampling_rate: int = 16000
    hop_length: int = 160
    chunk_length: int = 30
    n_fft: int = 400
    padding: str = "longest"
    padding_value: float = 0.0
    padding_side: str = "right"
    return_attention_mask: bool = False


@register_preprocessor("whisper_feature_extractor", config_class=WhisperFeatureExtractorConfig)
class WhisperFeatureExtractor(AudioFeatureExtractor):
    """
    A feature extractor for Whisper model.

    This feature extractor inherits from `AudioFeatureExtractor` which contains most of the main methods.

    This class extracts mel-filter bank features from raw speech using a custom numpy implementation of the `Short Time
    Fourier Transform` which should match pytorch's `torch.stft` equivalent.
    """

    def __init__(self, config: WhisperFeatureExtractorConfig, **kwargs):
        super().__init__(config=config, **kwargs)
        self.n_samples = self.config.chunk_length * self.config.sampling_rate
        self.nb_max_frames = self.n_samples // self.config.hop_length
        self.mel_filters = mel_filter_bank(
            num_frequency_bins=1 + self.config.n_fft // 2,
            num_mel_filters=self.config.feature_size,
            min_frequency=0.0,
            max_frequency=8000.0,
            sampling_rate=self.config.sampling_rate,
            norm="slaney",
            mel_scale="slaney",
        )

    def __call__(
        self,
        raw_speech: Union[np.ndarray, List[float], List[np.ndarray], List[List[float]]],
        device: str = None,
        truncation: bool = True,
        pad_to_multiple_of: int = None,
        return_tensors: str = None,
        return_attention_mask: bool = None,
        padding: str = "max_length",
        max_length: int = None,
        sampling_rate: int = None,
        do_normalize: bool = None,
        **kwargs,
    ):
        sampling_rate = sampling_rate or self.config.sampling_rate
        if sampling_rate is not None:
            if sampling_rate != self.config.sampling_rate:
                raise ValueError(
                    f"The model corresponding to this feature extractor: {self.__class__.__name__} was trained using a"
                    f" sampling rate of {self.config.sampling_rate}. Please make sure that the provided `raw_speech` "
                    f"input was sampled with {self.config.sampling_rate} and not {sampling_rate}."
                )
        else:
            logger.warning(
                "It is strongly recommended to pass the `sampling_rate` argument to this function. "
                "Failing to do so can result in silent errors that might be hard to debug."
            )

        is_batched_numpy = isinstance(raw_speech, np.ndarray) and len(raw_speech.shape) > 1
        if is_batched_numpy and len(raw_speech.shape) > 2:
            raise ValueError(f"Only mono-channel audio is supported for input to {self}")
        is_batched = is_batched_numpy or (
            isinstance(raw_speech, (list, tuple)) and (isinstance(raw_speech[0], (np.ndarray, tuple, list)))
        )

        if is_batched:
            raw_speech = [np.asarray([speech], dtype=np.float32).T for speech in raw_speech]
        elif not is_batched and not isinstance(raw_speech, np.ndarray):
            raw_speech = np.asarray(raw_speech, dtype=np.float32)
        elif isinstance(raw_speech, np.ndarray) and raw_speech.dtype is np.dtype(np.float64):
            raw_speech = raw_speech.astype(np.float32)

        # always return batch
        if not is_batched:
            raw_speech = [np.asarray([raw_speech]).T]

        batched_speech = {"input_features": raw_speech}

        padded_inputs = self.pad(
            batched_speech,
            padding=padding,
            max_length=max_length if max_length else self.n_samples,
            truncation=truncation,
            pad_to_multiple_of=pad_to_multiple_of,
            return_attention_mask=return_attention_mask or do_normalize,
            return_tensors="np",
        )

        # zero-mean and unit-variance normalization
        if do_normalize:
            padded_inputs["input_features"] = self.zero_mean_unit_var_norm(
                padded_inputs["input_features"],
                attention_mask=padded_inputs["attention_mask"],
                padding_value=self.config.padding_value,
            )
            padded_inputs["input_features"] = np.stack(padded_inputs["input_features"], axis=0)

        # make sure list is in array format
        input_features = padded_inputs.get("input_features").transpose(2, 0, 1)

        input_features = [self._np_extract_fbank_features(waveform) for waveform in input_features[0]]

        if isinstance(input_features[0], List):
            padded_inputs["input_features"] = [np.asarray(feature, dtype=np.float32) for feature in input_features]
        else:
            padded_inputs["input_features"] = input_features

        if return_attention_mask:
            # rescale from sample (48000) to feature (3000)
            padded_inputs["attention_mask"] = padded_inputs["attention_mask"][:, :: self.config.hop_length]

        if return_tensors is not None:
            padded_inputs = {k: np.asarray(v) for k, v in padded_inputs.items()}
            padded_inputs = convert_batch_dict_dtype(padded_inputs, dtype=return_tensors)
            if device:
                padded_inputs = {
                    k: v.to(device) if isinstance(v, torch.Tensor) else v for k, v in padded_inputs.items()
                }

        return padded_inputs

    def _np_extract_fbank_features(self, waveform: np.array) -> np.ndarray:
        """
        Compute the log-mel spectrogram of the provided audio, gives similar results to Whisper's original torch
        implementation with 1e-5 tolerance.
        """
        log_spec = spectrogram(
            waveform,
            window_function(self.config.n_fft, "hann"),
            frame_length=self.config.n_fft,
            hop_length=self.config.hop_length,
            power=2.0,
            mel_filters=self.mel_filters,
            log_mel="log10",
        )
        log_spec = log_spec[:, :-1]
        log_spec = np.maximum(log_spec, log_spec.max() - 8.0)
        log_spec = (log_spec + 4.0) / 4.0
        return log_spec

    @staticmethod
    def zero_mean_unit_var_norm(
        input_values: List[np.ndarray], attention_mask: List[np.ndarray], padding_value: float = 0.0
    ) -> List[np.ndarray]:
        """
        Every array in the list is normalized to have zero mean and unit variance
        """
        if attention_mask is not None:
            attention_mask = np.array(attention_mask, np.int32)
            normed_input_values = []

            for vector, length in zip(input_values, attention_mask.sum(-1)):
                normed_slice = (vector - vector[:length].mean()) / np.sqrt(vector[:length].var() + 1e-7)
                if length < normed_slice.shape[0]:
                    normed_slice[length:] = padding_value

                normed_input_values.append(normed_slice)
        else:
            normed_input_values = [(x - x.mean()) / np.sqrt(x.var() + 1e-7) for x in input_values]

        return normed_input_values



from .whisper_speech_recognition import WhisperSpeechRecognition
from .whisper_feature_extractor import WhisperFeatureExtractor, WhisperFeatureExtractorConfig
from .whisper_speech_recognition_config import WhisperSpeechRecognitionConfig
from .whisper_tokenizer import WhisperBPETokenizer, WhisperBPEConfig



from re import sub
from typing import Dict, List


__all__ = [
    "snake_case",
    "permute_dict_list",
]


def snake_case(s):
    return "_".join(sub("([A-Z][a-z]+)", r" \1", sub("([A-Z]+)", r" \1", s.replace("-", " "))).split()).lower()


def permute_dict_list(dict_list: List[Dict]) -> Dict[str, List]:
    """
    Convert a list of dictionaries to a dictionary of lists

    Args:
        dict_list: Input list of dicts

    Returns:

    """
    if not len(dict_list):
        return {}
    d = {key: [x[key] for x in dict_list] for key in dict_list[0]}
    return d



import numpy as np
import torch

from ..preprocessors import Tokenizer
from ..utils import Logger, convert_batch_dict_dtype


__all__ = [
    "TextPaddingDataCollator",
    "SequenceLabelingDataCollator",
    "CharLevelOCRDataCollator",
]

logger = Logger(__name__)


class TextPaddingDataCollator:
    """
    A data collator that pads a batch of tokenized inputs.

    Args:
        tokenizer: A Hezar tokenizer instance. (only its config is going to be used)
        padding_type: Specifies padding strategy. Defaults to `longest`, but can also be `max_length` (in this case
         `max_length` cannot be None)
        padding_side: Specifies from which side of each tensor to add paddings. Defaults to `right`, but can also be
         `left`.
        max_length: If `padding_type` is set to `max_length` this parameter must be specified. Forces all tensors to
         have this value as length.
        return_tensors: Specifies the dtype of the returning tensors in the batch. Defaults to `pt(torch.Tensor)`, but
         can also be `np` or `list`.

    """

    def __init__(
        self,
        tokenizer: Tokenizer,
        padding_type: str = "longest",
        padding_side: str = "right",
        max_length: int = None,
        return_tensors: str = "pt",
    ):
        self.tokenizer = tokenizer
        self.padding_type = padding_type
        self.padding_side = padding_side
        self.max_length = max_length
        self.return_tensors = return_tensors

        self.field_to_pad_id_mapping = {
            "token_ids": self.tokenizer.pad_token_id,
            "token_type_ids": self.tokenizer.config.pad_token_type_id,
            "tokens": "",
            "special_tokens_mask": 1,
            "attention_mask": 0,
        }

        if padding_type == "longest" and max_length is not None:
            logger.warning(
                "You passed `max_length` while also setting `padding_type` to `longest` which are "
                "incompatible! Instead leave `max_length` as None or set `padding_type` to `max_length`! "
                "Ignoring `max_length`"
            )
            self.max_length = None

    def __call__(self, encoded_batch):
        """
        Add padding to every item in the batch

        Args:
            encoded_batch: A batch dictionary

        Returns:
            The same batch dictionary but padded
        """
        encoded_batch = [convert_batch_dict_dtype(x, dtype="list") for x in encoded_batch]
        permuted_batch = {}
        for key in encoded_batch[0].keys():
            stack = [e for item in encoded_batch for e in item[key]]
            permuted_batch[key] = stack

        encoded_batch = permuted_batch.copy()
        if "label" in encoded_batch:
            encoded_batch["labels"] = encoded_batch["label"]
            del encoded_batch["label"]

        labels = encoded_batch.pop("labels")
        input_length = self.max_length or max(len(x) for x in encoded_batch["token_ids"])

        for field, batch in encoded_batch.items():
            padded_batch = []
            for x in batch:
                if isinstance(x, torch.Tensor):
                    x = x.cpu().numpy().tolist()
                elif isinstance(x, np.ndarray):
                    x = x.tolist()
                difference = input_length - len(x)
                paddings = [self.field_to_pad_id_mapping[field]] * difference
                padded_x = x + paddings if self.padding_side == "right" else paddings + x
                padded_batch.append(padded_x)
            encoded_batch[field] = padded_batch

        encoded_batch["labels"] = labels

        encoded_batch = convert_batch_dict_dtype(encoded_batch, dtype=self.return_tensors)

        return encoded_batch


class SequenceLabelingDataCollator:
    """
    A data collator for sequence labeling.

    Args:
        tokenizer: A Hezar tokenizer instance. (only its config is going to be used)
        padding_type: Specifies padding strategy. Defaults to `longest`, but can also be `max_length` (in this case
         `max_length` cannot be None)
        padding_side: Specifies from which side of each tensor to add paddings. Defaults to `right`, but can also be
         `left`.
        max_length: If `padding_type` is set to `max_length` this parameter must be specified. Forces all tensors to
         have this value as length.
        return_tensors: Specifies the dtype of the returning tensors in the batch. Defaults to `pt(torch.Tensor)`, but
         can also be `np` or `list`.

    """

    def __init__(
        self,
        tokenizer: Tokenizer,
        padding_type: str = "longest",
        padding_side: str = "right",
        label_pad_token_id: int = -100,
        max_length: int = None,
        return_tensors: str = "pt",
    ):
        self.tokenizer = tokenizer
        self.padding_type = padding_type
        self.padding_side = padding_side
        self.label_pad_token_id = label_pad_token_id
        self.max_length = max_length
        self.return_tensors = return_tensors

    def __call__(self, encoded_batch):
        label_name = "label" if "label" in encoded_batch[0].keys() else "labels"
        labels = [feature[label_name] for feature in encoded_batch] if label_name in encoded_batch[0].keys() else None
        self.tokenizer.config.padding_direction = self.padding_side
        batch = self.tokenizer.pad_encoded_batch(
            encoded_batch,
            padding=self.padding_type,  # noqa
            max_length=self.max_length,
            # Conversion to tensors will fail if we have labels as they are not of the same length yet.
            return_tensors="pt" if labels is None else None,
        )

        if labels is None:
            return batch

        batch.pop("word_ids")
        sequence_length = torch.tensor(batch["token_ids"]).shape[1]
        if self.padding_side == "right":
            batch[label_name] = [
                list(label) + [self.label_pad_token_id] * (sequence_length - len(label)) for label in labels
            ]
        else:
            batch[label_name] = [
                [self.label_pad_token_id] * (sequence_length - len(label)) + list(label) for label in labels
            ]

        batch = {k: torch.tensor(v, dtype=torch.int64) for k, v in batch.items()}
        return batch


class CharLevelOCRDataCollator:
    def __init__(self, pad_token_id: int = 0):
        self.pad_token_id = pad_token_id

    def __call__(self, input_batch):
        if isinstance(input_batch, (list, tuple)) and isinstance(input_batch[0], dict):
            input_batch = {key: [example[key] for example in input_batch] for key in input_batch[0].keys()}
        input_batch["pixel_values"] = torch.stack(input_batch["pixel_values"], 0)
        input_batch["labels_length"] = torch.cat(input_batch["labels_length"], 0)

        max_length = max(map(len, input_batch["labels"]))
        all_labels = []
        for labels in input_batch["labels"]:
            labels = labels.numpy().tolist()
            labels += [self.pad_token_id] * (max_length - len(labels))
            all_labels.append(labels)
        input_batch["labels"] = torch.tensor(all_labels)
        return input_batch



import os
import tempfile
from collections import defaultdict
from dataclasses import dataclass
from typing import Dict, List, Literal, Mapping, Optional, Tuple, Union

import numpy as np
import torch
from huggingface_hub import create_repo, hf_hub_download, upload_file

from ...builders import build_preprocessor
from ...configs import PreprocessorConfig
from ...constants import (
    DEFAULT_TOKENIZER_CONFIG_FILE,
    DEFAULT_TOKENIZER_FILE,
    HEZAR_CACHE_DIR,
    Backends,
)
from ...utils import Logger, convert_batch_dict_dtype, is_backend_available
from ..preprocessor import Preprocessor


if is_backend_available(Backends.TOKENIZERS):
    from tokenizers import Tokenizer as HFTokenizer
    from tokenizers.decoders import Decoder
    from tokenizers.models import Model

logger = Logger(__name__)


@dataclass
class TokenizerConfig(PreprocessorConfig):
    name = "tokenizer"
    max_length: int = None
    truncation_strategy: str = None
    truncation_direction: str = None
    stride: int = None
    padding_strategy: str = None
    padding_direction: str = None
    pad_to_multiple_of: int = None
    pad_token_type_id: int = 0
    bos_token: str = None
    eos_token: str = None
    unk_token: str = None
    sep_token: str = None
    pad_token: str = None
    cls_token: str = None
    mask_token: str = None
    additional_special_tokens: List[str] = None


class Tokenizer(Preprocessor):
    """
    Base tokenizer class. Mostly copied from :class: ~tokenizers.implementations.BaseTokenizer

    Args:
        config: A TokenizerConfig instance
        tokenizer_file: A tokenizer.json file to load the whole tokenizer from
        **kwargs: Extra config parameters that merge into the main config
    """

    required_backends: List[Union[str, Backends]] = []

    tokenizer_filename = DEFAULT_TOKENIZER_FILE
    tokenizer_config_filename = DEFAULT_TOKENIZER_CONFIG_FILE
    token_ids_name = "token_ids"
    uncastable_keys = ["word_ids", "tokens", "offsets_mapping"]

    def __init__(self, config: TokenizerConfig, tokenizer_file=None, **kwargs):
        super().__init__(config, **kwargs)
        self._tokenizer = self.from_file(tokenizer_file) if tokenizer_file is not None else self.build()
        self.special_tokens = self._get_all_special_tokens()

    def _get_all_special_tokens(self):
        _special_tokens = [
            self.config.bos_token,
            self.config.eos_token,
            self.config.unk_token,
            self.config.sep_token,
            self.config.pad_token,
            self.config.cls_token,
            self.config.mask_token,
        ]
        _special_tokens = [token for token in _special_tokens if token in self.vocab]

        if self.config.additional_special_tokens is not None:
            for token in self.config.additional_special_tokens:
                if token not in _special_tokens:
                    _special_tokens.append(token)

        valid_tokens = [token for token in _special_tokens if token is not None]
        return valid_tokens

    @staticmethod
    def from_file(path):
        tokenizer = HFTokenizer.from_file(path)
        return tokenizer

    def build(self):
        raise NotImplementedError

    def encode(self, inputs, is_pretokenized: bool = False, add_special_tokens: bool = True, **kwargs):
        """
        tokenize a list of inputs(could be raw or tokenized inputs)
        Args:
            inputs:
            is_pretokenized: Whether the inputs are already tokenized
            add_special_tokens: Whether to add special tokens to the inputs. Defaults to True.
            **kwargs:

        Returns:

        """
        if isinstance(inputs, str):
            inputs = [inputs]
        elif isinstance(inputs, list) and is_pretokenized:
            if isinstance(inputs[0], str):
                inputs = [inputs]
        return self._tokenizer.encode_batch(inputs, is_pretokenized, add_special_tokens)

    def decode(self, ids: List[int], skip_special_tokens: bool = True, **kwargs):
        if isinstance(ids[0], int):
            ids = [ids]
        if isinstance(ids, torch.Tensor):
            ids = ids.cpu().numpy().tolist()
        if isinstance(ids, np.ndarray):
            ids = ids.tolist()
        return self._tokenizer.decode_batch(ids, skip_special_tokens=skip_special_tokens)

    def pad_encoded_batch(
        self,
        inputs,
        padding: Literal["longest", "max_length"] = None,
        max_length: Optional[int] = None,
        return_tensors: Optional[str] = None,
        skip_keys: list = None,
    ):
        """
        Given a batch of encoded inputs, add padding to all of them so that the inputs are of the same length.

        Args:
            inputs: Input batch of encoded tokens
            padding: Padding type, could be one of ["longest", "max_length"]
            max_length: Max input length (only if padding is set to "max_length")
            return_tensors: The type of tensors to return
            skip_keys: A list of keys to skip padding

        Returns:

        """
        if isinstance(inputs, (list, tuple)) and isinstance(inputs[0], Mapping):
            inputs = {key: [example[key] for example in inputs] for key in inputs[0].keys()}

        inputs_max_length = max([len(x) for x in inputs[self.token_ids_name]])
        # resolve padding and max_length parameters
        padding = padding or self.config.padding_strategy

        if padding is None:
            if max_length is not None:
                padding = "max_length"
            elif max_length is None:
                logger.warning("Both padding and max_length are None so the inputs cannot be padded!")
                return inputs

        if padding == "longest":
            if max_length is not None:
                logger.warning(
                    "Setting padding='longest' and max_length is not valid. You must set one of them"
                    "and leave the other as None. Falling back to padding='longest'"
                )

            inputs_length = inputs_max_length

        elif padding == "max_length":
            if max_length is None:
                logger.warning(
                    "Setting padding='max_length' but no max_length value is provided in the function "
                    "parameters nor the tokenizer config! Falling back to padding='longest'"
                )
                inputs_length = inputs_max_length
            else:
                # TODO implement truncation if possible and remove this condition
                if max_length <= inputs_max_length:
                    logger.warning(
                        f"Cannot set max_length to {max_length} "
                        f"while max input length is {inputs_max_length}!"
                        f"Falling back to padding='longest' "
                        f"since truncation is not available yet in Hezar :("
                    )
                    inputs_length = inputs_max_length
                else:
                    inputs_length = max_length

        skip_keys = skip_keys or []
        skip_keys += self.uncastable_keys  # avoid possible errors
        inputs = convert_batch_dict_dtype(inputs, dtype="list", skip_keys=skip_keys)

        for key, batch in inputs.items():
            if key in skip_keys:
                continue
            padding_id = 0 if key == "attention_mask" else self.pad_token_id
            padded_batch = []
            for x in batch:
                difference = inputs_length - len(x)
                paddings = [padding_id] * difference
                padded_x = x + paddings if self.config.padding_direction == "right" else paddings + x
                padded_batch.append(padded_x)
            inputs[key] = padded_batch

        inputs = convert_batch_dict_dtype(inputs, dtype=return_tensors, skip_keys=skip_keys)

        return inputs

    def __call__(
        self,
        inputs: Union[List[str], List[Tuple[str, str]]],
        device: Union[str, torch.device] = None,
        add_special_tokens: bool = True,
        padding_strategy=None,
        truncation_strategy=None,
        max_length: int = None,
        return_tensors: str = "list",
        stride: int = 0,
        is_split_into_words: bool = False,
        pad_to_multiple_of: int = None,
        return_tokens: bool = None,
        return_token_type_ids: bool = None,
        return_attention_mask: bool = True,
        return_overflowing_tokens: bool = False,
        return_special_tokens_mask: bool = True,
        return_offsets_mapping: bool = False,
        return_length: bool = False,
        return_word_ids: bool = False,
        verbose: bool = True,
        **kwargs,
    ):
        """
        Tokenize a batch of string inputs and return the relevant properties e.g, token ids, attention mask, etc.

        Args:
            inputs: A list of string inputs to tokenize
            add_special_tokens: Whether to add special tokens or not
            padding_strategy: Determines how to pad inputs
            truncation_strategy: Determines how to truncate inputs
            max_length: Max input length of the sequences
            return_tensors: The type of the returning tensors in the batch e.g, pt, np, list
            stride: Stride level
            is_split_into_words: Are inputs pre-tokenized or raw string inputs
            pad_to_multiple_of: Pad inputs by a factor of this value
            return_tokens: Whether to return tokens lists
            return_token_type_ids: Whether to return token type ids
            return_attention_mask: Whether to return attention masks
            return_overflowing_tokens: Whether to return overflowing tokens
            return_special_tokens_mask: Whether to return special tokens mask
            return_offsets_mapping: Whether to return offsets
            return_length: Whether to return input lengths
            **kwargs: Extra arguments reside here and therefore ignored

        Returns:
            A dictionary of encoded inputs like
                {"token_ids": [batch_size x input_len], "attention_mask": [batch_size x input_len], ...}
        """
        if isinstance(inputs, list) and not len(inputs):
            raise ValueError("Tokenizer cannot process an empty list!")

        if isinstance(inputs, str):
            inputs = [inputs]

        padding_strategy = padding_strategy or self.config.padding_strategy
        truncation_strategy = truncation_strategy or self.config.truncation_strategy
        max_length = max_length or self.config.max_length
        pad_to_multiple_of = pad_to_multiple_of or self.config.pad_to_multiple_of

        self.set_truncation_and_padding(
            padding_strategy=padding_strategy,
            truncation_strategy=truncation_strategy,
            padding_side=self.config.padding_direction,
            truncation_side=self.config.truncation_direction,
            max_length=max_length,
            stride=self.config.stride,
            pad_to_multiple_of=pad_to_multiple_of,
        )
        encodings = self.encode(
            inputs,
            add_special_tokens=add_special_tokens,
            is_pretokenized=is_split_into_words,
        )
        encodings_dict = [
            self._convert_encodings(
                encoding=encoding,
                return_tokens=return_tokens,
                return_token_type_ids=return_token_type_ids,
                return_attention_mask=return_attention_mask,
                return_overflowing_tokens=return_overflowing_tokens,
                return_special_tokens_mask=return_special_tokens_mask,
                return_offsets_mapping=return_offsets_mapping,
                return_length=return_length,
                return_word_ids=return_word_ids,
            )
            for encoding in encodings
        ]
        # Permute output dict from [batch_0: Dict[key, value], ...] to Dict[key, [batch_0, batch_1, ...], ...]
        sanitized_outputs = {}
        for key in encodings_dict[0].keys():
            stack = [e for item in encodings_dict for e in item[key]]
            sanitized_outputs[key] = stack

        # If returning overflowing tokens, we need to return a mapping
        # from the batch idx to the original sample
        if return_overflowing_tokens:
            overflow_to_sample_mapping = []
            for i, encodings_ in enumerate(encodings_dict):
                overflow_to_sample_mapping += [i] * len(encodings_["input_ids"])
            sanitized_outputs["overflow_to_sample_mapping"] = overflow_to_sample_mapping

        if return_tensors == "list" or return_tensors is None:
            sanitized_outputs = {
                key: value[0] if len(value) > 0 and isinstance(value[0], list) else value
                for key, value in sanitized_outputs.items()
            }

        outputs = convert_batch_dict_dtype(sanitized_outputs, dtype=return_tensors, skip_keys=self.uncastable_keys)
        if device and return_tensors == "pt":
            outputs = {k: v.to(device) if isinstance(v, torch.Tensor) else v for k, v in outputs.items()}

        return outputs

    def set_truncation_and_padding(
        self,
        padding_strategy=None,
        truncation_strategy=None,
        padding_side=None,
        truncation_side=None,
        max_length: int = None,
        stride: int = None,
        pad_to_multiple_of: int = None,
    ):
        # Set truncation and padding on the backend tokenizer
        if truncation_strategy == "no_truncation":
            if self.truncation is not None:
                self.no_truncation()
        else:
            target = {
                "max_length": max_length,
                "stride": stride,
                "strategy": truncation_strategy,
                "direction": truncation_side,
            }
            if self.truncation is None:
                current = None
            else:
                current = {k: self.truncation.get(k, None) for k in target}

            if current != target:
                self.enable_truncation(**target)

            if padding_strategy == "no_padding":
                if self.padding is not None:
                    self.no_padding()
            else:
                length = max_length if self.config.padding_strategy == "max_length" else None
                target = {
                    "length": length,
                    "direction": padding_side,
                    "pad_id": self.token_to_id(self.pad_token),
                    "pad_token": self.pad_token,
                    "pad_type_id": self.config.pad_token_type_id,
                    "pad_to_multiple_of": pad_to_multiple_of,
                }
                if self.padding != target:
                    self.enable_padding(**target)

    def _convert_encodings(
        self,
        encoding,
        return_tokens: bool = None,
        return_token_type_ids: bool = None,
        return_attention_mask: bool = None,
        return_overflowing_tokens: bool = False,
        return_special_tokens_mask: bool = False,
        return_offsets_mapping: bool = False,
        return_length: bool = False,
        return_word_ids: bool = False,
    ):
        if return_overflowing_tokens and encoding.overflowing is not None:
            encodings = [encoding] + encoding.overflowing
        else:
            encodings = [encoding]

        encoding_dict = defaultdict(list)
        for e in encodings:
            encoding_dict[self.token_ids_name].append(e.ids)

            if return_token_type_ids:
                encoding_dict["token_type_ids"].append(e.type_ids)
            if return_attention_mask:
                encoding_dict["attention_mask"].append(e.attention_mask)
            if return_special_tokens_mask:
                encoding_dict["special_tokens_mask"].append(e.special_tokens_mask)
            if return_offsets_mapping:
                encoding_dict["offsets_mapping"].append(e.offsets)
            if return_length:
                encoding_dict["length"].append(len(e.ids))
            if return_tokens:
                text = self._tokenizer.decode(e.ids)
                tokens = self.get_tokens_from_offsets(text, e.ids, e.offsets)
                encoding_dict["tokens"].append(tokens)
            if return_word_ids:
                encoding_dict["word_ids"].append(e.word_ids)

        return encoding_dict

    def convert_tokens_to_ids(self, tokens: Union[str, List[str]]) -> Union[int, List[int]]:
        if isinstance(tokens, str):
            tokens = [tokens]

        return [self._tokenizer.token_to_id(token) for token in tokens]

    def convert_ids_to_tokens(self, ids: Union[int, List[int]], skip_special_tokens: bool = False):
        if isinstance(ids, int):
            ids = [ids]
        tokens = []
        for index in ids:
            index = int(index)
            if skip_special_tokens and index in self.special_ids:
                continue
            tokens.append(self._tokenizer.id_to_token(index))
        return tokens

    def num_special_tokens_to_add(self, is_pair: bool) -> int:
        return self._tokenizer.num_special_tokens_to_add(is_pair)

    def get_vocab(self, with_added_tokens: bool = True) -> Dict[str, int]:
        return self._tokenizer.get_vocab(with_added_tokens=with_added_tokens)

    def get_vocab_size(self, with_added_tokens: bool = True) -> int:
        return self._tokenizer.get_vocab_size(with_added_tokens=with_added_tokens)

    def enable_padding(
        self,
        direction: str = "right",
        pad_to_multiple_of: int = None,
        pad_id: int = 0,
        pad_type_id: int = 0,
        pad_token: str = "[PAD]",
        length: int = None,
    ):
        return self._tokenizer.enable_padding(
            direction=direction,
            pad_to_multiple_of=pad_to_multiple_of,
            pad_id=pad_id,
            pad_type_id=pad_type_id,
            pad_token=pad_token,
            length=length,
        )

    def no_padding(self):
        return self._tokenizer.no_padding()

    def enable_truncation(self, max_length, stride=0, strategy="longest_first", direction="right"):
        return self._tokenizer.enable_truncation(max_length, stride=stride, strategy=strategy, direction=direction)

    def no_truncation(self):
        return self._tokenizer.no_truncation()

    def add_tokens(self, tokens) -> int:
        return self._tokenizer.add_tokens(tokens)

    def add_special_tokens(self, special_tokens) -> int:
        return self._tokenizer.add_special_tokens(special_tokens)

    def token_to_id(self, token: str) -> int:
        return self._tokenizer.token_to_id(token)

    def id_to_token(self, id: int) -> str:
        return self._tokenizer.id_to_token(id)

    def get_added_vocab(self) -> Dict[str, int]:
        """
        Returns the added tokens in the vocabulary as a dictionary of token to index.

        Returns:
            `Dict[str, int]`: The added tokens.
        """
        base_vocab = self._tokenizer.get_vocab(with_added_tokens=False)
        full_vocab = self._tokenizer.get_vocab(with_added_tokens=True)
        added_vocab = {token: index for token, index in full_vocab.items() if token not in base_vocab}
        return added_vocab

    def __len__(self) -> int:
        """
        Size of the full vocabulary with the added tokens.
        """
        return self._tokenizer.get_vocab_size(with_added_tokens=True)

    def get_tokens_from_offsets(
        self,
        text: Union[str, List[str]],
        ids: List[int],
        offsets_mapping: List[Tuple[int, int]],
    ):
        """
        Extract human-readable tokens using the original text and offsets mapping
        Args:
            text: Raw string text
            ids: Token ids
            offsets_mapping: A list of tuples representing offsets

        Returns:
            A list of tokens
        """
        if not isinstance(text, str):
            raise ValueError(f"Expected str type for `text`, got `{type(text)}({text})`")
        if isinstance(offsets_mapping, list) and not isinstance(offsets_mapping[0], Tuple):
            raise ValueError(f"Expected a list of tuples for `offsets_mapping`, got List[{type(offsets_mapping[0])}]")
        tokens = []
        for offset in offsets_mapping:
            offset_start, offset_end = offset
            tokens.append(text[offset_start:offset_end])
        for i, token in enumerate(tokens):
            if ids[i] in self.special_ids:
                tokens[i] = self._tokenizer.id_to_token(ids[i])
        return tokens

    @classmethod
    def load(
        cls,
        hub_or_local_path,
        subfolder=None,
        config_filename=None,
        tokenizer_filename=None,
        **kwargs,
    ) -> "Tokenizer":
        tokenizer_filename = tokenizer_filename or cls.tokenizer_filename
        config_filename = config_filename or cls.tokenizer_config_filename
        subfolder = subfolder or cls.preprocessor_subfolder
        config = TokenizerConfig.load(
            hub_or_local_path,
            filename=config_filename,
            subfolder=subfolder,
        )

        if os.path.isdir(hub_or_local_path):
            tokenizer_path = os.path.join(hub_or_local_path, subfolder, tokenizer_filename)
        else:
            tokenizer_path = hf_hub_download(
                hub_or_local_path,
                filename=tokenizer_filename,
                subfolder=subfolder,
                cache_dir=HEZAR_CACHE_DIR,
                resume_download=True,
            )
        tokenizer = build_preprocessor(config.name, config, tokenizer_file=tokenizer_path, **kwargs)
        return tokenizer

    def save(self, path, save_config=True, pretty=True):
        os.makedirs(path, exist_ok=True)
        # save config
        if save_config:
            self.config.vocab_size = self.get_vocab_size(with_added_tokens=True)
            self.config.save(path, filename=self.tokenizer_config_filename, subfolder=self.preprocessor_subfolder)
        # save tokenizer.json
        save_path = os.path.join(path, self.preprocessor_subfolder, self.tokenizer_filename)
        self._tokenizer.save(save_path, pretty=pretty)

    def push_to_hub(
        self,
        repo_id,
        commit_message=None,
        subfolder=None,
        tokenizer_filename=None,
        config_filename=None,
        private=False,
    ):
        """
        Push tokenizer and config to the Hub

        Args:
            repo_id: The path (id or repo name) on the hub
            commit_message: Commit message for this push
            subfolder: subfolder to save the files
            tokenizer_filename: tokenizer filename
            config_filename: tokenizer config filename
            private: If the repo should be private (ignored if the repo exists)
        """
        subfolder = subfolder or self.preprocessor_subfolder
        tokenizer_filename = tokenizer_filename or self.tokenizer_filename
        config_filename = config_filename or self.tokenizer_config_filename

        # create remote repo
        create_repo(repo_id, exist_ok=True, private=private)
        # save to tmp and prepare for push
        cache_path = tempfile.mkdtemp()
        # save tokenizer.json
        tokenizer_save_path = os.path.join(cache_path, subfolder, tokenizer_filename)
        self.save(cache_path, pretty=True)

        if commit_message is None:
            commit_message = "Hezar: Upload tokenizer and config"

        # upload config
        self.config.push_to_hub(
            repo_id=repo_id,
            filename=config_filename,
            subfolder=subfolder,
            commit_message=commit_message,
        )
        # upload tokenizer
        upload_file(
            repo_id=repo_id,
            path_or_fileobj=tokenizer_save_path,
            repo_type="model",
            path_in_repo=f"{subfolder}/{tokenizer_filename}",
            commit_message=commit_message,
        )
        logger.log_upload_success(
            name=f"{self.__class__.__name__}(name={self.config.name})",
            target_path=os.path.join(repo_id, subfolder, tokenizer_filename),
        )

    @property
    def model(self) -> "Model":
        return self._tokenizer.model

    @model.setter
    def model(self, model: "Model"):
        self._tokenizer.model = model  # noqa

    @property
    def decoder(self) -> "Decoder":
        return self._tokenizer.decoder

    @decoder.setter
    def decoder(self, decoder: "Decoder"):
        self._tokenizer.decoder = decoder  # noqa

    @property
    def padding(self):
        return self._tokenizer.padding

    @property
    def truncation(self) -> dict:
        return self._tokenizer.truncation

    @property
    def vocab(self):
        return self._tokenizer.get_vocab(with_added_tokens=True)

    @property
    def vocab_size(self) -> int:
        """
        `int`: Size of the base vocabulary (without the added tokens).
        """
        return self._tokenizer.get_vocab_size(with_added_tokens=False)

    @property
    def special_ids(self):
        return [self.token_to_id(t) for t in self.special_tokens]

    @property
    def pad_token(self):
        return self.config.pad_token

    @property
    def bos_token(self):
        return self.config.bos_token

    @property
    def eos_token(self):
        return self.config.eos_token

    @property
    def unk_token(self):
        return self.config.unk_token

    @property
    def mask_token(self):
        return self.config.mask_token

    @property
    def cls_token(self):
        return self.config.cls_token

    @property
    def sep_token(self):
        return self.config.sep_token

    @property
    def pad_token_id(self):
        return self.token_to_id(self.config.pad_token)

    @property
    def bos_token_id(self):
        return self.token_to_id(self.config.bos_token)

    @property
    def eos_token_id(self):
        return self.token_to_id(self.config.eos_token)

    @property
    def unk_token_id(self):
        return self.token_to_id(self.config.unk_token)

    @property
    def mask_token_id(self):
        return self.token_to_id(self.config.mask_token)

    @property
    def cls_token_id(self):
        return self.token_to_id(self.config.cls_token)

    @property
    def sep_token_id(self):
        return self.token_to_id(self.config.sep_token)



import logging


__all__ = ["Logger"]


class Logger(logging.Logger):
    def __init__(self, name: str, level=None, fmt=None):
        fmt = fmt or "Hezar (%(levelname)s): %(message)s"
        level = level or "INFO"
        super().__init__(name, level)
        handler = logging.StreamHandler()
        formatter = logging.Formatter(fmt)
        handler.setFormatter(formatter)
        self.addHandler(handler)

    def log_upload_success(self, name, target_path: str):
        """
        Log (info) success info when the file(s) upload is done.
        """
        self.info(f"Uploaded: `{name}` --> `{target_path}`")



"""
A DISTILBERT model for sequence labeling built using HuggingFace Transformers
"""
from typing import Dict, List, Union

import torch
from torch import nn

from ....constants import Backends
from ....registry import register_model
from ....utils import is_backend_available
from ...model import Model
from ...model_outputs import SequenceLabelingOutput
from .distilbert_sequence_labeling_config import DistilBertSequenceLabelingConfig


if is_backend_available(Backends.TRANSFORMERS):
    from transformers import DistilBertConfig, DistilBertModel

_required_backends = [
    Backends.TRANSFORMERS,
    Backends.TOKENIZERS,
]


@register_model("distilbert_sequence_labeling", DistilBertSequenceLabelingConfig)
class DistilBertSequenceLabeling(Model):
    required_backends = _required_backends
    tokenizer_name = "wordpiece_tokenizer"

    def __init__(self, config: DistilBertSequenceLabelingConfig, **kwargs):
        super().__init__(config, **kwargs)
        self.distilbert = DistilBertModel(self._build_inner_config())
        classifier_dropout = (
            config.classifier_dropout if config.classifier_dropout is not None else config.hidden_dropout_prob
        )
        self.dropout = nn.Dropout(classifier_dropout)
        self.classifier = nn.Linear(config.dim, config.num_labels)

    def _build_inner_config(self):
        """
        Build the inner config for DistilBert. If `num_labels` is not provided, it will be inferred from `id2label`.
        If only `num_labels` is provided, `id2label` will be inferred from `num_labels` using the default label names.
        :return:
        """
        if self.config.num_labels is None and self.config.id2label is None:
            raise ValueError("Both `num_labels` and `id2label` are None. Please provide at least one of them!")
        if self.config.id2label is not None and self.config.num_labels is None:
            self.config.num_labels = len(self.config.id2label)
        config = DistilBertConfig(**self.config)
        return config

    def forward(
        self,
        token_ids,
        attention_mask=None,
        head_mask=None,
        inputs_embeds=None,
        output_attentions=None,
        output_hidden_states=None,
        **kwargs,
    ) -> Dict:
        lm_outputs = self.distilbert(
            input_ids=token_ids,
            attention_mask=attention_mask,
            head_mask=head_mask,
            inputs_embeds=inputs_embeds,
            output_attentions=output_attentions,
            output_hidden_states=output_hidden_states,
        )
        sequence_output = lm_outputs[0]

        sequence_output = self.dropout(sequence_output)
        logits = self.classifier(sequence_output)

        outputs = {
            "logits": logits,
            "hidden_states": lm_outputs.hidden_states,
            "attentions": lm_outputs.attentions,
            "tokens": kwargs.get("tokens", None),
            "offsets": kwargs.get("offsets_mapping", None)
        }
        return outputs

    def compute_loss(self, logits: torch.Tensor, labels: torch.Tensor) -> torch.Tensor:
        criterion = nn.CrossEntropyLoss()
        loss = criterion(logits.view(-1, self.config.num_labels), labels.view(-1))
        return loss

    def preprocess(self, inputs: Union[str, List[str]], **kwargs):
        if isinstance(inputs, str):
            inputs = [inputs]
        if "text_normalizer" in self.preprocessor:
            normalizer = self.preprocessor["text_normalizer"]
            inputs = normalizer(inputs)
        tokenizer = self.preprocessor[self.tokenizer_name]
        inputs = tokenizer(
            inputs,
            return_word_ids=True,
            return_tokens=True,
            return_offsets_mapping=True,
            padding=True,
            truncation=True,
            return_tensors="pt",
            device=self.device,
        )
        return inputs

    def post_process(
        self,
        model_outputs: Dict[str, torch.Tensor],
        return_offsets: bool = False,
        return_scores: bool = False,
    ):
        logits = model_outputs["logits"].softmax(2)
        tokens = model_outputs["tokens"]
        offsets = model_outputs["offsets"]
        probs, predictions = logits.max(2)
        predictions = [[self.config.id2label[p.item()] for p in prediction] for prediction in predictions]
        outputs = []
        for tokens_list, prediction, probs_, offsets_mapping in zip(tokens, predictions, probs, offsets):
            results = []
            for token, label, prob, offset in zip(tokens_list, prediction, probs_, offsets_mapping):
                if token not in self.config.prediction_skip_tokens:
                    token_results = {"token": token, "label": label}
                    if return_scores:
                        token_results["score"] = prob.item()
                    if return_offsets:
                        start, end = offset
                        token_results["start"] = start
                        token_results["end"] = end
                    results.append(SequenceLabelingOutput(**token_results))
            outputs.append(results)
        return outputs



from dataclasses import dataclass, field
from typing import List

from ...constants import DEFAULT_TOKENIZER_CONFIG_FILE, DEFAULT_TOKENIZER_FILE, Backends
from ...registry import register_preprocessor
from ...utils import is_backend_available
from .tokenizer import Tokenizer, TokenizerConfig


if is_backend_available(Backends.TOKENIZERS):
    from tokenizers import Tokenizer as HFTokenizer
    from tokenizers import decoders, models, pre_tokenizers, trainers

_required_backends = [
    Backends.TOKENIZERS,
]


@dataclass
class SentencePieceUnigramConfig(TokenizerConfig):
    name = "sentencepiece_unigram_tokenizer"
    max_length: int = 512
    truncation_strategy: str = "longest_first"
    truncation_direction: str = "right"
    stride: int = 0
    padding_strategy: str = "longest"
    padding_direction: str = "right"
    bos_token: str = "<s>"
    eos_token: str = "</s>"
    unk_token: str = "<unk>"
    sep_token: str = "<sep>"
    pad_token: str = "<pad>"
    cls_token: str = "<cls>"
    mask_token: str = "<mask>"
    additional_special_tokens: List[str] = None
    pad_to_multiple_of: int = 0
    dropout: float = None
    continuing_subword_prefix: str = ""
    replacement: str = "_"
    add_prefix_space: bool = True
    end_of_word_suffix: str = ""
    fuse_unk: bool = False
    vocab_size: int = 8000
    min_frequency: int = 2
    limit_alphabet: int = 1000
    initial_alphabet: list = field(default_factory=list)
    show_progress: bool = True


@register_preprocessor("sentencepiece_unigram_tokenizer", config_class=SentencePieceUnigramConfig)
class SentencePieceUnigramTokenizer(Tokenizer):
    """
    A standard SentencePiece Unigram tokenizer using 🤗HuggingFace Tokenizers

    Args:
        config: Preprocessor config for the tokenizer
        **kwargs: Extra/manual config parameters
    """

    required_backends = _required_backends

    tokenizer_filename = DEFAULT_TOKENIZER_FILE
    tokenizer_config_filename = DEFAULT_TOKENIZER_CONFIG_FILE
    token_ids_name = "token_ids"

    def __init__(self, config, tokenizer_file=None, **kwargs):
        super().__init__(config, tokenizer_file=tokenizer_file, **kwargs)

    def build(self):
        tokenizer = HFTokenizer(models.Unigram())  # noqa
        tokenizer.normalizer = normalizers.Sequence(  # noqa
            [  # noqa
                normalizers.Nmt(),  # noqa
                normalizers.NFKC(),  # noqa
                normalizers.Replace(Regex(" {2,}"), " "),  # noqa
            ]  # noqa
        )  # noqa
        tokenizer.pre_tokenizer = pre_tokenizers.Metaspace(  # noqa
            replacement=self.config.replacement, add_prefix_space=self.config.add_prefix_space
        )
        tokenizer.decoder = decoders.Metaspace(  # noqa
            replacement=self.config.replacement, add_prefix_space=self.config.add_prefix_space
        )

        return tokenizer

    def train(self, files: List[str], **train_kwargs):
        """Train the model using the given files"""
        self.config.update(train_kwargs)

        trainer = trainers.UnigramTrainer(
            vocab_size=self.config.vocab_size,  # noqa
            special_tokens=self.config.special_tokens,  # noqa
            show_progress=self.config.show_progress,  # noqa
            initial_alphabet=self.config.initial_alphabet,  # noqa
            unk_token=self.config.unk_token,  # noqa
        )
        if isinstance(files, str):
            files = [files]
        self._tokenizer.train(files, trainer=trainer)

    def train_from_iterator(self, dataset: List[str], **train_kwargs):
        """Train the model using the given files"""
        self.config.update(train_kwargs)

        trainer = trainers.UnigramTrainer(
            vocab_size=self.config.vocab_size,  # noqa
            special_tokens=self.config.special_tokens,  # noqa
            show_progress=self.config.show_progress,  # noqa
            initial_alphabet=self.config.initial_alphabet,  # noqa
            unk_token=self.config.unk_token,  # noqa
        )
        self._tokenizer.train_from_iterator(dataset, trainer=trainer, length=len(dataset))



"""
Hezar models inherit the base class `Model`. A `Model` itself is a PyTorch Module to implement neural networks but has
some extra Hezar-specific functionalities and methods e.g, pushing to hub, loading from hub, etc.

Examples:
    >>> # Load from hub
    >>> from hezar import Model
    >>> model = Model.load("hezarai/bert-base-fa")
"""
import inspect
import os
import tempfile
from collections import OrderedDict
from typing import Any, Dict, Iterable, List, Mapping, Optional, Union

import torch
from huggingface_hub import create_repo, hf_hub_download, upload_file
from torch import nn

from ..builders import build_model
from ..configs import ModelConfig
from ..constants import (
    DEFAULT_MODEL_CONFIG_FILE,
    DEFAULT_MODEL_FILE,
    HEZAR_CACHE_DIR,
    Backends,
    LossType,
    RegistryType,
)
from ..preprocessors import Preprocessor, PreprocessorsContainer
from ..utils import Logger, get_module_class, verify_dependencies
from .model_outputs import ModelOutput


logger = Logger(__name__)

criterions_mapping = {
    "l1": nn.L1Loss,
    "nll": nn.NLLLoss,
    "nll_2d": nn.NLLLoss2d,
    "poisson_nll": nn.PoissonNLLLoss,
    "gaussian_nll": nn.GaussianNLLLoss,
    "mse": nn.MSELoss,
    "bce": nn.BCELoss,
    "bce_with_logits": nn.BCEWithLogitsLoss,
    "cross_entropy": nn.CrossEntropyLoss,
    "triple_margin": nn.TripletMarginLoss,
    "ctc": nn.CTCLoss
}


class Model(nn.Module):
    """
    Base class for all neural network models in Hezar.

    Args:
        config: A dataclass model config
    """

    required_backends: List[Union[Backends, str]] = []
    # Default file names
    model_filename = DEFAULT_MODEL_FILE
    config_filename = DEFAULT_MODEL_CONFIG_FILE

    # Specify if the model is a generative model. If True, the model must also implement the `generate` method
    is_generative: bool = False

    # Keys to ignore on loading state dicts
    skip_keys_on_load = []

    # Loss function name
    loss_fn_name: Union[str, LossType] = LossType.CROSS_ENTROPY

    def __init__(self, config: ModelConfig, *args, **kwargs):
        verify_dependencies(self, self.required_backends)
        super().__init__()
        self.config = config.update(kwargs)
        self._preprocessor = None
        self._criterion = self._set_criterion(self.loss_fn_name)

    @staticmethod
    def _set_criterion(criterion_name: str):
        if criterion_name not in criterions_mapping:
            raise ValueError(f"Invalid criterion name `{criterion_name}`. Available: {list(criterions_mapping.keys())}")
        loss_fn = criterions_mapping[criterion_name]()
        return loss_fn

    @classmethod
    def load(
        cls,
        hub_or_local_path: Union[str, os.PathLike],
        load_locally: Optional[bool] = False,
        load_preprocessor: Optional[bool] = True,
        model_filename: Optional[str] = None,
        config_filename: Optional[str] = None,
        save_path: Optional[Union[str, os.PathLike]] = None,
        **kwargs,
    ) -> "Model":
        """
        Load the model from local path or hub.

        It's recommended to actually use this method with :class:`hezar.Model` rather than any other model class
        unless you actually know that the class is the same as the one on the Hub, because the output will always be
        the one specified on the Hub!

        Args:
            hub_or_local_path: Path to the model living on the Hub or local disk.
            load_locally: Force loading from local path
            load_preprocessor: Whether to load the preprocessor(s) or not
            model_filename: Optional model filename.
            config_filename: Optional config filename
            save_path: Save model to this path after loading

        Returns:
            The fully loaded Hezar model
        """
        # Get device if provided in the kwargs
        device = None or kwargs.pop("device", None)
        # Load config
        config_filename = config_filename or cls.config_filename
        config = ModelConfig.load(hub_or_local_path=hub_or_local_path, filename=config_filename)
        # Get the exact model class based on registry name under config.name
        model_cls = get_module_class(config.name, registry_type=RegistryType.MODEL)
        # Handle compatibility of model class and the one in the config
        if cls.__name__ == "Model":
            # Build model wih config
            model = build_model(config.name, config, **kwargs)
        else:
            if cls.__name__ != model_cls.__name__:
                logger.warning(
                    f"You attempted to load a model using `{cls.__name__}` "
                    f"but the model in `{hub_or_local_path}` is of type `{model_cls.__name__}`, "
                    f"so the output model will be of type {model_cls.__name__} anyway!"
                )
            model = model_cls(config, **kwargs)

        model_filename = model_filename or model_cls.model_filename or cls.model_filename
        # does the path exist locally?
        is_local = load_locally or os.path.isdir(hub_or_local_path)
        if not is_local:
            model_path = hf_hub_download(
                hub_or_local_path,
                filename=model_filename,
                cache_dir=HEZAR_CACHE_DIR,
                resume_download=True,
            )
        else:
            model_path = os.path.join(hub_or_local_path, model_filename)
        # Get state dict from the model
        state_dict = torch.load(model_path, map_location=torch.device("cpu"))
        model.load_state_dict(state_dict)
        if device:
            model.to(device)
        if save_path:
            model.save(save_path)
        # Load the preprocessor(s)
        if load_preprocessor:
            preprocessor = Preprocessor.load(hub_or_local_path, force_return_dict=True)
            model.preprocessor = preprocessor
        return model

    def load_state_dict(self, state_dict: Mapping[str, Any], **kwargs):
        """
        Flexibly load the state dict to the model.

        Any incompatible or missing key is ignored and other layer weights are
        loaded. In that case a warning with additional info is raised.

        Args:
            state_dict: Model state dict
        """
        if len(self.skip_keys_on_load):
            for key in self.skip_keys_on_load:
                if key in state_dict:
                    state_dict.pop(key, None)  # noqa
        try:
            super().load_state_dict(state_dict, strict=True)
        except RuntimeError:
            compatible_state_dict = OrderedDict()
            src_state_dict = self.state_dict()

            incompatible_keys = []

            for (src_key, src_weight), (trg_key, trg_weight) in zip(src_state_dict.items(), state_dict.items()):
                if src_weight.shape == trg_weight.shape:
                    compatible_state_dict[src_key] = trg_weight
                else:
                    # put the source key and weight if trg weight is incompatible
                    compatible_state_dict[src_key] = src_weight
                    incompatible_keys.append(src_key)

            missing_keys, _ = super().load_state_dict(compatible_state_dict, strict=False)
            if len(missing_keys) or len(incompatible_keys):
                logger.warning(
                    "Partially loading the weights as the model architecture and the given state dict are "
                    "incompatible! \nIgnore this warning in case you plan on fine-tuning this model\n"
                    f"Incompatible keys: {incompatible_keys}\n"
                    f"Missing keys: {missing_keys}\n"
                )

    def save(
        self,
        path: Union[str, os.PathLike],
        filename: Optional[str] = None,
        save_preprocessor: Optional[bool] = True,
        config_filename: Optional[str] = None,
    ):
        """
        Save model weights and config to a local path

        Args:
            path: A local directory to save model, config, etc.
            save_preprocessor: Whether to save preprocessor(s) along with the model or not
            config_filename: Model config filename,
            filename: Model weights filename

        Returns:
            Path to the saved model
        """
        # save model and config to the repo
        config_filename = config_filename or self.config_filename
        filename = filename or self.model_filename
        os.makedirs(path, exist_ok=True)

        self.config.save(save_dir=path, filename=config_filename)

        model_save_path = os.path.join(path, filename)
        torch.save(self.state_dict(), model_save_path)

        if save_preprocessor:
            if self.preprocessor is not None:
                self.preprocessor.save(path)

    def push_to_hub(
        self,
        repo_id: str,
        filename: Optional[str] = None,
        config_filename: Optional[str] = None,
        push_preprocessor: Optional[bool] = True,
        commit_message: Optional[str] = None,
        private: Optional[bool] = False,
    ):
        """
        Push the model and required files to the hub

        Args:
            repo_id: The path (id or repo name) on the hub
            filename: Model file name
            config_filename: Config file name
            push_preprocessor: Whether to push preprocessor(s) or not
            commit_message (str): Commit message for this push
            private (bool): Whether to create a private repo or not
        """
        config_filename = config_filename or self.config_filename
        filename = filename or self.model_filename

        # create remote repo
        create_repo(repo_id, repo_type="model", exist_ok=True, private=private)

        # save to tmp and prepare for push
        cache_path = tempfile.mkdtemp()
        self.save(cache_path, filename=filename, config_filename=config_filename)
        if commit_message is None:
            commit_message = "Hezar: Upload model and config"

        # upload config file
        self.config.push_to_hub(
            repo_id,
            filename=config_filename,
            repo_type="model",
            commit_message=commit_message,
        )
        # upload preprocessor(s)
        if push_preprocessor:
            if self.preprocessor is not None:
                self.preprocessor.push_to_hub(repo_id, commit_message=commit_message, private=private)

        # upload model file
        weights_path = os.path.join(cache_path, filename)
        upload_file(
            path_or_fileobj=weights_path,
            path_in_repo=filename,
            repo_id=repo_id,
            commit_message=commit_message,
        )

        logger.log_upload_success(
            name=f"{self.__class__.__name__}(name={self.config.name})",
            target_path=os.path.join(repo_id, filename),
        )

    def forward(self, *model_inputs, **kwargs) -> Dict:
        """
        Forward inputs through the model and return logits, etc.

        Args:
            model_inputs: The required inputs for the model forward

        Returns:
            A dict of outputs like logits, loss, etc.
        """
        raise NotImplementedError

    def compute_loss(self, inputs: torch.Tensor, targets: torch.Tensor) -> torch.Tensor:
        """
        Compute loss on the model outputs against the given labels

        Args:
            inputs: Input tensor to compute loss on
            targets: Target tensor

        Returns:
            Loss tensor
        """
        raise NotImplementedError

    def generate(self, *model_inputs, **kwargs) -> torch.Tensor:
        """
        Generation method for all generative models. Generative models have the `is_generative` attribute set to True.
        The behavior of this method is usually controlled by `generation` part of the model's config.

        Args:
            model_inputs: Model inputs for generation, usually the same as forward's `model_inputs`
            **kwargs: Generation kwargs

        Returns:
            Generated output tensor
        """
        raise NotImplementedError

    def preprocess(self, *raw_inputs: Union[Any, List[Any]], **kwargs):
        """
        Given raw inputs, preprocess the inputs and prepare them for model's `forward()`.

        Args:
            raw_inputs: Raw model inputs
            **kwargs: Extra kwargs specific to the model. See the model's specific class for more info

        Returns:
            A dict of inputs for model forward
        """
        return raw_inputs

    def post_process(self, *model_outputs: Union[Dict[str, torch.Tensor], torch.Tensor], **kwargs):
        """
        Process model outputs and return human-readable results. Called in `self.predict()`

        Args:
            model_outputs: model outputs to process
            **kwargs: extra arguments specific to the derived class

        Returns:
            Processed model output values and converted to human-readable results
        """
        return model_outputs

    @torch.inference_mode()
    def predict(
        self,
        inputs: Union[Any, List[Any]],
        device: Union[str, torch.device] = None,
        unpack_forward_inputs: bool = True,
        **kwargs,
    ) -> Union[Dict, List[Dict], torch.Tensor, Iterable, ModelOutput]:
        """
        Perform an end-to-end prediction on raw inputs.

        If the model is a generative model, it has to implement the `generate()` method too which will be called
        instead of `forward()`. (`forward()` method is called internally within the `generate()` method)

        Args:
            inputs: Raw inputs e.g, a list of texts, path to images, etc.
            device: What device to perform inference on
            unpack_forward_inputs: Whether to unpack forward inputs. Set to False if you want to send preprocess outputs
             directly to the forward/generate method without unpacking it. Note that this only applies to the cases that
             the preprocess method's output is a dict-like/mapping object.
            **kwargs: Other arguments for `preprocess`, `forward`, `generate` and `post_process`. each will be passed to
             the correct method automatically.

        Returns:
            Output dict of results
        """
        # Unpack kwargs for each step
        preprocess_kwargs, forward_kwargs, post_process_kwargs = self._unpack_prediction_kwargs(**kwargs)
        invalid_kwargs = {
            k: v for k, v in kwargs.items() if k not in {**preprocess_kwargs, **forward_kwargs, **post_process_kwargs}
        }
        if len(invalid_kwargs):
            logger.warning(
                f"Unrecognized arguments {list(invalid_kwargs.keys())} passed to `predict` method for "
                f"`{self.__class__.__name__}`"
            )

        # Put model in eval mode
        self.eval()

        # Preprocessing step
        model_inputs = self.preprocess(inputs, **preprocess_kwargs)

        # Map inputs and model to device
        device = device or self.device
        model_inputs = self._move_inputs_to_device(model_inputs, device)
        self.to(device)

        # Specify model inference function
        inference_fn = self.generate if self.is_generative else self.__call__

        # Model inference step (forward for regular models and generate for generative models)
        if isinstance(model_inputs, Mapping) and unpack_forward_inputs:
            model_outputs = inference_fn(**model_inputs, **forward_kwargs)
        else:
            model_outputs = inference_fn(model_inputs, **forward_kwargs)

        # Post-processing step
        processed_outputs = self.post_process(model_outputs, **post_process_kwargs)
        return processed_outputs

    @staticmethod
    def _move_inputs_to_device(inputs, device):
        """
        Move all input tensors in the inputs to the device

        Args:
            inputs: A torch.Tensor or a batch dict that contains tensors in its values
            device: A torch compatible device

        Returns:
            Same inputs moved to the device
        """
        if isinstance(inputs, torch.Tensor):
            inputs = inputs.to(device)
        elif isinstance(inputs, Mapping):
            inputs = {k: v.to(device) if isinstance(v, torch.Tensor) else v for k, v in inputs.items()}
        else:
            raise ValueError(
                f"Cannot move inputs of type `{type(inputs)}` to `{device}`. "
                f"Inputs data type must be either `torch.Tensor` or a mapping object like `dict`!"
            )
        return inputs

    def _unpack_prediction_kwargs(self, **kwargs):
        """
        The `predict` method can accept extra parameters for each of the `preprocess`, `forward/generate`
        and `post_process` methods. These parameters are passed as keyword arguments so that we have to make sure that
        each of them are passed to the correct method.

        Args:
            **kwargs: The kwargs to be unpacked

        Returns:
             A 3-sized tuple of (preprocess_kwargs, forward_kwargs, post_process_kwargs)
        """
        # Whether to use forward or generate based on model type
        inference_fn = type(self).generate if self.is_generative else type(self).forward

        def _get_positional_kwargs(fn):
            params = dict(inspect.signature(fn).parameters)
            params = {k: v for k, v in params.items() if v.default != v.empty}
            return params

        # Get keyword arguments from the child class (ignore positional arguments)
        preprocess_kwargs_keys = list(_get_positional_kwargs(type(self).preprocess).keys())
        post_process_kwargs_keys = list(_get_positional_kwargs(type(self).post_process).keys())
        forward_kwargs_keys = list(_get_positional_kwargs(inference_fn).keys())

        preprocess_kwargs = {k: kwargs.get(k) for k in preprocess_kwargs_keys if k in kwargs}
        forward_kwargs = {k: kwargs.get(k) for k in forward_kwargs_keys if k in kwargs}
        post_process_kwargs = {k: kwargs.get(k) for k in post_process_kwargs_keys if k in kwargs}

        return preprocess_kwargs, forward_kwargs, post_process_kwargs

    @property
    def device(self):
        """
        Get the model's device. This method is only safe when all weights of the model are on the same device.
        """
        return next(self.parameters()).device

    @property
    def criterion(self):
        return self._criterion

    @criterion.setter
    def criterion(self, value):
        if isinstance(value, str):
            self._criterion = self._set_criterion(value)
        elif isinstance(value, nn.Module):
            self._criterion = value
        else:
            raise ValueError(f"Criterion value must be either a name or a PyTorch `nn.Module`, got {type(value)}!")

    @property
    def preprocessor(self) -> PreprocessorsContainer:
        return self._preprocessor

    @preprocessor.setter
    def preprocessor(self, value: Union[Preprocessor, PreprocessorsContainer, List[Preprocessor]]):
        """
        A safe setter method for model's preprocessor. Value must be either a Preprocessor, a list of Preprocessors or
        a PreprocessorsContainer instance.
        """
        if isinstance(value, Preprocessor):
            preprocessor = PreprocessorsContainer()
            preprocessor[value.config.name] = value
        elif isinstance(value, Mapping):
            preprocessor = PreprocessorsContainer(**value)
        elif isinstance(value, list):
            preprocessor = PreprocessorsContainer(**{p.config.name: p for p in value})
        elif value is None:
            preprocessor = None
        else:
            raise ValueError(
                f"Preprocessor value must be a `Preprocessor` "
                f"or a list of Preprocessor objects"
                f"or `PreprocessorContainer` instance not `{type(value)}`!"
            )
        self._preprocessor = preprocessor



from dataclasses import dataclass, field

from ....configs import ModelConfig


@dataclass
class DecoderConfig(ModelConfig):
    name = "vit_gpt2_decoder"
    add_cross_attention: bool = True
    vocab_size: int = 42001
    attn_pdrop: float = 0.1
    bos_token_id: int = 5
    embd_pdrop: float = 0.1
    eos_token_id: int = 5
    gradient_checkpointing: bool = False
    initializer_range: float = 0.02
    layer_norm_epsilon: float = 1e-05
    model_type: str = "gpt2"
    n_ctx: int = 1024
    n_embd: int = 768
    n_head: int = 12
    n_inner: int = None
    n_layer: int = 12
    n_positions: int = 1024
    resid_pdrop: float = 0.1
    summary_activation: bool = False
    summary_first_dropout: float = 0.1
    use_cache: bool = True


@dataclass
class EncoderConfig(ModelConfig):
    name = "vit_gpt2_encoder"
    hidden_size: int = 768
    num_hidden_layers: int = 12
    num_attention_heads: int = 12
    intermediate_size: int = 3072
    hidden_act: str = "gelu"
    hidden_dropout_prob: float = 0.0
    attention_probs_dropout_prob: float = 0.0
    initializer_range: float = 0.02
    layer_norm_eps: float = 1e-12
    image_size: int = 224
    patch_size: int = 16
    num_channels: int = 3
    qkv_bias: bool = True
    encoder_stride: int = 16


@dataclass
class GenerationConfig(ModelConfig):
    bos_token_id: int = 0
    decoder_start_token_id: int = 0
    early_stopping: bool = True
    eos_token_id: int = 2
    length_penalty: float = 2.0
    max_new_tokens: int = 24
    no_repeat_ngram_size: int = 3
    num_beams: int = 4
    pad_token_id: int = 1


@dataclass
class ViTGPT2Image2TextConfig(ModelConfig):
    name = "vit_gpt2_image2text"
    encoder: EncoderConfig = field(default_factory=EncoderConfig)
    decoder: DecoderConfig = field(default_factory=DecoderConfig)
    generation: GenerationConfig = field(default_factory=GenerationConfig)



from .bert import BertTextClassification, BertTextClassificationConfig
from .distilbert import DistilBertTextClassification, DistilBertTextClassificationConfig
from .roberta import RobertaTextClassification, RobertaTextClassificationConfig



import importlib.util
from importlib.metadata import version
from typing import List, Union

from ..constants import Backends


__all__ = [
    "is_backend_available",
    "verify_dependencies",
    "get_lib_version",
]


def is_backend_available(backend: Backends):
    """
    Check if the backend package is installed or not

    Args:
        backend: Package name

    Returns:
        Whether the package is available or not
    """
    return importlib.util.find_spec(backend) is not None


def verify_dependencies(obj, backends: List[Union[Backends, str]] = None):
    """
    Check if all the required dependencies are installed or not.

    Args:
        obj: The target object to check. (Usually `self`)
        backends: A list of dependency names of type `str` or `Backends`

    Raises:
        ModuleNotFoundError
    """
    if backends is None:
        return
    unavailable = []
    for backend in backends:
        if not is_backend_available(backend):
            unavailable.append(backend)
    if len(unavailable):
        raise ModuleNotFoundError(
            f"`{obj.__class__.__name__}` requires "
            f"{f'`{unavailable[0]}`' if len(unavailable) == 1 else unavailable} "
            f"which {'is' if len(unavailable) == 1 else 'are'} not installed!"
        )


def get_lib_version(lib: str):
    return version(lib)



from .audio_utils import *
from .common_utils import *
from .context_managers import *
from .core_utils import *
from .data_utils import *
from .file_utils import *
from .hub_utils import *
from .image_utils import *
from .integration_utils import *
from .logging import *
from .registry_utils import *



from dataclasses import dataclass

from ..configs import MetricConfig
from ..constants import Backends, MetricType
from ..registry import register_metric
from ..utils import is_backend_available
from .metric import Metric


if is_backend_available(Backends.JIWER):
    import jiwer
    import jiwer.transforms as tr

_DESCRIPTION = "Character Error Rate (CER) using `jiwer`. Commonly used for Speech Recognition and OCR systems"

_required_backends = [
    Backends.JIWER,
]


@dataclass
class CERConfig(MetricConfig):
    name = MetricType.CER
    sentence_delimiter: str = " "
    concatenate_texts: bool = False
    output_keys: tuple = ("cer",)


@register_metric("cer", config_class=CERConfig, description=_DESCRIPTION)
class CER(Metric):
    required_backends = _required_backends

    def __init__(self, config: CERConfig, **kwargs):
        super().__init__(config=config, **kwargs)
        self.transform = tr.Compose(
            [
                tr.RemoveMultipleSpaces(),
                tr.Strip(),
                tr.ReduceToSingleSentence(self.config.sentence_delimiter),
                tr.ReduceToListOfListOfChars(),
            ]
        )

    def compute(
        self,
        predictions=None,
        targets=None,
        concatenate_texts=None,
        n_decimals=None,
        output_keys=None,
        **kwargs,
    ):
        concatenate_texts = concatenate_texts or self.config.concatenate_texts
        n_decimals = n_decimals or self.config.n_decimals

        if concatenate_texts:
            score = jiwer.compute_measures(
                targets,
                predictions,
                truth_transform=self.transform,
                hypothesis_transform=self.transform,
            )["wer"]

        else:
            incorrect = 0
            total = 0
            for prediction, reference in zip(predictions, targets):
                measures = jiwer.compute_measures(
                    reference,
                    prediction,
                    truth_transform=self.transform,
                    hypothesis_transform=self.transform,
                )
                incorrect += measures["substitutions"] + measures["deletions"] + measures["insertions"]
                total += measures["substitutions"] + measures["deletions"] + measures["hits"]

            score = incorrect / total

        results = {"cer": round(float(score), n_decimals)}

        if output_keys:
            results = {k: v for k, v in results.items() if k in output_keys}

        return results



"""
Home to all constant variables in Hezar
"""

import os
from enum import Enum


HEZAR_HUB_ID = "hezarai"
HEZAR_CACHE_DIR = os.getenv("HEZAR_CACHE_DIR", f'{os.path.expanduser("~")}/.cache/hezar')

DEFAULT_MODEL_FILE = "model.pt"
DEFAULT_MODEL_CONFIG_FILE = "model_config.yaml"
DEFAULT_TRAINER_SUBFOLDER = "train"
DEFAULT_TRAINER_CONFIG_FILE = "train_config.yaml"
DEFAULT_TRAINER_CSV_LOG_FILE = "training_logs.csv"
DEFAULT_PREPROCESSOR_SUBFOLDER = "preprocessor"
DEFAULT_NORMALIZER_CONFIG_FILE = "normalizer_config.yaml"
DEFAULT_IMAGE_PROCESSOR_CONFIG_FILE = "image_processor_config.yaml"
DEFAULT_FEATURE_EXTRACTOR_CONFIG_FILE = "feature_extractor_config.yaml"
DEFAULT_TOKENIZER_FILE = "tokenizer.json"
DEFAULT_TOKENIZER_CONFIG_FILE = "tokenizer_config.yaml"
DEFAULT_DATASET_CONFIG_FILE = "dataset_config.yaml"
DEFAULT_EMBEDDING_FILE = "embedding.bin"
DEFAULT_EMBEDDING_CONFIG_FILE = "embedding_config.yaml"
DEFAULT_EMBEDDING_SUBFOLDER = "embedding"

TQDM_BAR_FORMAT = "{desc:<16}{percentage:3.0f}%|{bar:70}{r_bar}"


class Backends(str, Enum):
    """
    All required dependency packages and libraries. Note that the values here must be the exact module names used
    for importing, for example if you set PILLOW the value must be `PIL` not `pillow`, `pil`, etc.
    """

    PYTORCH = "torch"
    TRANSFORMERS = "transformers"
    DATASETS = "datasets"
    TOKENIZERS = "tokenizers"
    SOUNDFILE = "soundfile"
    LIBROSA = "librosa"
    WANDB = "wandb"
    GENSIM = "gensim"
    PILLOW = "PIL"
    JIWER = "jiwer"
    NLTK = "nltk"
    SCIKIT = "sklearn"
    SEQEVAL = "seqeval"

    def __str__(self):
        return str(self.value)


class TaskType(str, Enum):
    AUDIO_CLASSIFICATION = "audio_classification"
    BACKBONE = "backbone"
    IMAGE2TEXT = "image2text"
    LANGUAGE_MODELING = "language_modeling"
    SEQUENCE_LABELING = "sequence_labeling"
    SPEECH_RECOGNITION = "speech_recognition"
    TEXT_CLASSIFICATION = "text_classification"
    TEXT_DETECTION = "text_detection"
    TEXT_GENERATION = "text_generation"


class ConfigType(str, Enum):
    BASE = "base"
    MODEL = "model"
    DATASET = "dataset"
    PREPROCESSOR = "preprocessor"
    EMBEDDING = "embedding"
    TRAINER = "trainer"
    OPTIMIZER = "optimizer"
    CRITERION = "criterion"
    LR_SCHEDULER = "lr_scheduler"
    METRIC = "metric"


class RegistryType(str, Enum):
    MODEL = "model"
    DATASET = "dataset"
    PREPROCESSOR = "preprocessor"
    EMBEDDING = "embedding"
    TRAINER = "trainer"
    OPTIMIZER = "optimizer"
    CRITERION = "criterion"
    LR_SCHEDULER = "lr_scheduler"
    METRIC = "metric"


class LossType(str, Enum):
    L1 = "l1"
    NLL = "nll"
    NLL_2D = "nll_2d"
    POISSON_NLL = "poisson_nll"
    GAUSSIAN_NLL = "gaussian_nll"
    MSE = "mse"
    BCE = "bce"
    BCE_WITH_LOGITS = "bce_with_logits"
    CROSS_ENTROPY = "cross_entropy"
    TRIPLE_MARGIN = "triple_margin"
    CTC = "ctc"


class SplitType(str, Enum):
    TRAIN = "train"
    EVAL = "eval"
    VALID = "validation"
    TEST = "test"


class MetricType(str, Enum):
    ACCURACY = "accuracy"
    F1 = "f1"
    RECALL = "recall"
    PRECISION = "precision"
    SEQEVAL = "seqeval"
    CER = "cer"
    WER = "wer"
    BLEU = "bleu"


class RepoType(str, Enum):
    DATASET = "dataset"
    MODEL = "model"


class ImageType(str, Enum):
    NUMPY = "numpy"
    PILLOW = "pillow"
    TORCH = "torch"


class ChannelsAxisSide(str, Enum):
    FIRST = "first"
    LAST = "last"



"""
A RoBERTa Language Model (HuggingFace Transformers) wrapped by a Hezar Model class
"""
from typing import List, Union

import torch
from torch import nn, tanh

from ....constants import Backends
from ....registry import register_model
from ....utils import is_backend_available
from ...model import Model
from ...model_outputs import TextClassificationOutput
from .roberta_text_classification_config import RobertaTextClassificationConfig


if is_backend_available(Backends.TRANSFORMERS):
    from transformers import RobertaConfig, RobertaModel

_required_backends = [
    Backends.TRANSFORMERS,
    Backends.TOKENIZERS,
]


@register_model("roberta_text_classification", config_class=RobertaTextClassificationConfig)
class RobertaTextClassification(Model):
    """
    A standard 🤗Transformers RoBERTa model for text classification

    Args:
        config: The whole model config including arguments needed for the inner 🤗Transformers model.
    """

    required_backends = _required_backends
    tokenizer_name = "bpe_tokenizer"
    skip_keys_on_load = ["model.embeddings.position_ids", "roberta.embeddings.position_ids"]  # For older versions

    def __init__(self, config, **kwargs):
        super().__init__(config=config, **kwargs)
        self.roberta = RobertaModel(self._build_inner_config(), add_pooling_layer=False)
        self.classifier = RobertaClassificationHead(self.config)

    def _build_inner_config(self):
        if self.config.num_labels is None and self.config.id2label is None:
            raise ValueError("Both `num_labels` and `id2label` are None. Please provide at least one of them!")
        if self.config.id2label and self.config.num_labels is None:
            self.config.num_labels = len(self.config.id2label)
        bert_config = RobertaConfig(**self.config)
        return bert_config

    def forward(
        self,
        token_ids,
        attention_mask=None,
        token_type_ids=None,
        position_ids=None,
        head_mask=None,
        inputs_embeds=None,
        output_attentions=None,
        output_hidden_states=None,
        return_dict=None,
        **kwargs,
    ):
        lm_outputs = self.roberta(
            token_ids,
            attention_mask=attention_mask,
            token_type_ids=token_type_ids,
            position_ids=position_ids,
            head_mask=head_mask,
            inputs_embeds=inputs_embeds,
            output_attentions=output_attentions,
            output_hidden_states=output_hidden_states,
            return_dict=return_dict,
        )
        sequence_output = lm_outputs[0]
        logits = self.classifier(sequence_output)

        outputs = {
            "logits": logits,
            "hidden_states": lm_outputs.hidden_states,
            "attentions": lm_outputs.attentions,
        }
        return outputs

    def compute_loss(self, inputs: torch.Tensor, targets: torch.Tensor) -> torch.Tensor:
        criterion = nn.CrossEntropyLoss()
        loss = criterion(inputs.view(-1, self.config.num_labels), targets.view(-1))
        return loss

    def preprocess(self, inputs: Union[str, List[str]], **kwargs):
        if isinstance(inputs, str):
            inputs = [inputs]
        if "text_normalizer" in self.preprocessor:
            normalizer = self.preprocessor["text_normalizer"]
            inputs = normalizer(inputs)
        tokenizer = self.preprocessor[self.tokenizer_name]
        inputs = tokenizer(inputs, return_tensors="pt", device=self.device)
        return inputs

    def post_process(self, model_outputs: dict, top_k=1):
        output_logits = model_outputs["logits"]
        outputs = []
        for logits in output_logits:
            probs = logits.softmax(dim=-1)
            scores, label_ids = probs.sort(descending=True)
            row = []
            for i, (score, label_id) in enumerate(zip(scores, label_ids)):
                if i == top_k:
                    break
                label_str = self.config.id2label[label_id.item()]
                score = score.item()
                row.append(TextClassificationOutput(label=label_str, score=score))
            outputs.append(row)
        return outputs


class RobertaClassificationHead(nn.Module):
    """Head for sentence-level classification tasks."""

    def __init__(self, config):
        super().__init__()
        self.dense = nn.Linear(config.hidden_size, config.hidden_size)
        classifier_dropout = (
            config.classifier_dropout if config.classifier_dropout is not None else config.hidden_dropout_prob
        )
        self.dropout = nn.Dropout(classifier_dropout)
        self.out_proj = nn.Linear(config.hidden_size, config.num_labels)

    def forward(self, inputs, **kwargs):
        x = inputs[:, 0, :]  # take <s> token (equiv. to [CLS])
        x = self.dropout(x)
        x = self.dense(x)
        x = tanh(x)
        x = self.dropout(x)
        x = self.out_proj(x)
        return x



import os.path

from huggingface_hub import HfApi, Repository

from ..constants import HEZAR_CACHE_DIR, HEZAR_HUB_ID, RepoType
from ..utils.logging import Logger


__all__ = [
    "resolve_pretrained_path",
    "get_local_cache_path",
    "exists_in_cache",
    "exists_on_hub",
    "clone_repo",
    "list_repo_files",
]

logger = Logger(__name__)


def resolve_pretrained_path(hub_or_local_path):
    """
    **DEPRECATED**

    Resolve a local or Hub path. If path exists locally it just returns the input, otherwise tries to resolve
    hub_or_local_path. If it contains the namespace (author/org) leave it as is, otherwise change to hezarai/{hub_path}

    Args:
        hub_or_local_path: Repo name or id

    Returns:
        A proper pretrained path
    """
    logger.warning("`resolve_pretrained_path` is deprecated! Use the raw `hub_or_local_path`!")
    if os.path.isdir(hub_or_local_path):
        return hub_or_local_path
    repo_id = f"{HEZAR_HUB_ID}/{hub_or_local_path}" if "/" not in hub_or_local_path else hub_or_local_path
    return repo_id


def get_local_cache_path(repo_id, repo_type):
    """
    Given the hub path and repo type, configure the local path to save everything e.g, ~/.hezar/models/<repo_name>

    Args:
        repo_id: Repo name or id
        repo_type: Repo type e.g, model, dataset, etc

    Returns:
        Path to local cache directory
    """
    repo_owner, repo_name = repo_id.split("/")
    cache_path = f"{HEZAR_CACHE_DIR}/{repo_type}s--{repo_owner}--{repo_name}"
    return cache_path


def exists_in_cache(hub_path, repo_type="model"):
    cache_path = get_local_cache_path(hub_path, repo_type)
    return os.path.exists(cache_path)


def exists_on_hub(hub_path: str, repo_type="model"):
    """
    Determine whether the repo exists on the hub or not

    Args:
        hub_path: Repo name or id
        repo_type: Repo type like model, dataset, etc.

    Returns:
        True or False
    """
    author, repo_name = hub_path.split("/")
    api = HfApi()
    if repo_type == "model":
        paths = list(iter(api.list_models(author=author)))
    elif repo_type == "dataset":
        paths = list(iter(api.list_datasets(author=author)))
    elif repo_type == "space":
        paths = list(iter(api.list_spaces(author=author)))
    else:
        raise ValueError(f"Unknown type: {repo_type}! Use `model`, `dataset`, `space`, etc.")

    return hub_path in [path.id for path in paths]


def clone_repo(repo_id: str, save_path: str, **kwargs):
    """
    Clone a repo on the hub to local directory

    Args:
        repo_id: Repo name or id
        save_path: Path to clone the repo to

    Returns:
        the local path to the repo
    """
    repo = Repository(local_dir=save_path, clone_from=repo_id, **kwargs)
    return repo.local_dir


def list_repo_files(hub_or_local_path: str, subfolder: str = None):
    """
    List all files in a Hub or local model repo

    Args:
        hub_or_local_path: Path to hub or local repo
        subfolder: Optional subfolder path

    Returns:
        A list of all file names
    """
    if os.path.isdir(hub_or_local_path):
        files_itr = os.walk(hub_or_local_path)
        files = []
        for r, d, f in files_itr:
            if r == hub_or_local_path:
                files.append(f)
            else:
                for x in f:
                    files.append(f"{r.replace(f'{hub_or_local_path}/', '')}/{x}")
    else:
        files = HfApi().list_repo_files(hub_or_local_path, repo_type=RepoType.MODEL.value)

    if subfolder is not None:
        files = [x.replace(f"{subfolder}/", "") for x in files if subfolder in x]

    return files



r"""
Hezar uses a registry system in a way that for any core module like model, dataset, etc. there is an entry in its
specific registry. These registries are simple python dictionaries that map a module's name to its class and its config
class. These registries are initialized here and filled automatically when you import hezar or a registry itself.

Examples:
    >>> # read models registry
    >>> from hezar.registry import models_registry
    >>> print(models_registry)
    {'distilbert_lm': {'module_class': <class 'hezar.models.language_modeling.distilbert.distilbert_lm.DistilBertLM'>,
    'config_class': <class 'hezar.models.language_modeling.distilbert.distilbert_lm_config.DistilBertLMConfig'>},
    'description': 'Optional model description here...'}

    >>> # add a model class to models_registry
    >>> from hezar import Model, register_model
    >>> @register_model(name="my_awesome_model", config_class=MyAwesomeModelConfig, description="My Awesome Model!")
    >>> class MyAwesomeModel(Model):
    ...    def __init__(config: MyAwesomeModelConfig):
    ...        ...

Keep in mind that registries usually don't need to be used directly. There is a bunch of functions to build modules
using a module's registry name in `hezar.builders` module. See the file `builders.py` for more info.

Note: In case of adding a new registry container, make sure to add to `__all__` below!
"""

from dataclasses import dataclass
from typing import Dict, Optional, Type

from .configs import (
    DatasetConfig,
    EmbeddingConfig,
    MetricConfig,
    ModelConfig,
    PreprocessorConfig,
)
from .utils import Logger


__all__ = [
    "register_model",
    "register_preprocessor",
    "register_dataset",
    "register_embedding",
    "register_metric",
    "Registry",
    "models_registry",
    "preprocessors_registry",
    "datasets_registry",
    "embeddings_registry",
    "metrics_registry",
]

logger = Logger(__name__)


@dataclass
class Registry:
    module_class: type
    config_class: type = None
    description: Optional[str] = None


models_registry: Dict[str, Registry] = {}
preprocessors_registry: Dict[str, Registry] = {}
datasets_registry: Dict[str, Registry] = {}
embeddings_registry: Dict[str, Registry] = {}
metrics_registry: Dict[str, Registry] = {}


def register_model(model_name: str, config_class: Type[ModelConfig], description: str = None):
    """
    A class decorator that adds the model class and the config class to the `models_registry`

    Args:
        model_name: Model's registry name e.g, `bert_sequence_labeling`
        config_class: Model's config class e.g, `BertSequenceLabelingConfig`. This parameter must be the config class
            itself not a config instance!
        description: Optional model description
    """

    def register(cls):
        if model_name in models_registry:
            logger.warning(f"Model `{model_name}` is already registered. Overwriting...")

        if config_class.name != model_name:
            raise ValueError(
                f"`model_name` and `config.name` are not compatible for `{cls.__name__}`\n"
                f"model_name: {model_name}\n"
                f"{config_class.__name__}.name: {config_class.name}"
            )

        models_registry[model_name] = Registry(module_class=cls, config_class=config_class, description=description)

        return cls

    return register


def register_dataset(dataset_name: str, config_class: Type[DatasetConfig], description: str = None):
    """
    A class decorator that adds the dataset class and the config class to the `datasets_registry`

    Args:
        dataset_name: Dataset's registry name e.g, `text_classification`.
        config_class: Dataset's config class e.g, `TextClassificationDatasetConfig`. This parameter must be the config
            class itself not a config instance!
        description: Optional dataset description
    """

    def register(cls):
        if dataset_name in datasets_registry:
            logger.warning(f"Dataset `{dataset_name}` is already registered. Overwriting...")

        if config_class.name != dataset_name:
            raise ValueError(
                f"`dataset_name` and `config.name` are not compatible for `{cls.__name__}`\n"
                f"dataset_name: {dataset_name}\n"
                f"{config_class.__name__}.name: {config_class.name}"
            )

        datasets_registry[dataset_name] = Registry(
            module_class=cls,
            config_class=config_class,
            description=description,
        )

        return cls

    return register


def register_preprocessor(preprocessor_name: str, config_class: Type[PreprocessorConfig], description: str = None):
    """
    A class decorator that adds the preprocessor class and the config class to the `preprocessors_registry`

    Args:
        preprocessor_name: Preprocessor's registry name e.g, `bpe_tokenizer`.
        config_class: Preprocessor's config class e.g, BPEConfig. This parameter must be the config
            class itself not a config instance!
        description: Optional preprocessor description
    """

    def register(cls):
        if preprocessor_name in preprocessors_registry:
            logger.warning(f"Preprocessor `{preprocessor_name}` is already registered. Overwriting...")

        if config_class.name != preprocessor_name:
            raise ValueError(
                f"`preprocessor_name` and `config.name` are not compatible for `{cls.__name__}`\n"
                f"preprocessor_name: {preprocessor_name}\n"
                f"{config_class.__name__}.name: {config_class.name}"
            )

        preprocessors_registry[preprocessor_name] = Registry(
            module_class=cls,
            config_class=config_class,
            description=description,
        )

        return cls

    return register


def register_embedding(embedding_name: str, config_class: Type[EmbeddingConfig], description: str = None):
    """
    A class decorator that adds the embedding class and the config class to the `embeddings_registry`

    Args:
        embedding_name: Embedding's registry name e.g, `word2vec_cbow`.
        config_class: Embedding's config class e.g, Word2VecCBOWConfig. This parameter must be the config
            class itself not a config instance!
        description: Optional embedding description
    """

    def register(cls):
        if embedding_name in embeddings_registry:
            logger.warning(f"Embedding `{embedding_name}` is already registered. Overwriting...")

        if config_class.name != embedding_name:
            raise ValueError(
                f"`embedding_name` and `config.name` are not compatible for `{cls.__name__}`\n"
                f"embedding_name: {embedding_name}\n"
                f"{config_class.__name__}.name: {config_class.name}"
            )

        embeddings_registry[embedding_name] = Registry(
            module_class=cls,
            config_class=config_class,
            description=description,
        )

        return cls

    return register


def register_metric(metric_name: str, config_class: Type[MetricConfig], description: str = None):
    """
    A class decorator that adds the metric class and the config class to the `metrics_registry`

    Args:
        metric_name: Metric registry name e.g, `f1`
        config_class: Metric config class
        description: Optional metric description
    """

    def register(cls):
        if metric_name in metrics_registry:
            logger.warning(f"Metric `{metric_name}` is already registered. Overwriting...")
        if config_class.name != metric_name:
            raise ValueError(
                f"`metric_name` and `config.name` are not compatible for `{cls.__name__}`\n"
                f"metric_name: {metric_name}\n"
                f"{config_class.__name__}.name: {config_class.name}"
            )

        metrics_registry[metric_name] = Registry(
            module_class=cls,
            config_class=config_class,
            description=description,
        )

        return cls

    return register



from typing import Any, Dict

from .logging import Logger


__all__ = [
    "convert_batch_dict_dtype",
    "get_non_numeric_keys",
]

logger = Logger(__name__)


# TODO: This code might be able to be written in a cleaner way, but be careful, any change might break a lot of things!
def convert_batch_dict_dtype(batch_dict: Dict[str, Any], dtype: str = None, skip_keys: list = None):
    """
    Convert data dtypes of the values in a batch dict

    Args:
        batch_dict: The batched dictionary. Each key in the dict has a batch of data as its value.
        dtype: Target data type to convert to
        skip_keys: A list of key names to skip conversion

    Returns:
        The same dict with cast values
    """
    import numpy as np
    import torch

    dtype = dtype or "list"
    skip_keys = skip_keys or []

    if dtype == "list":
        for k, v in batch_dict.items():
            if isinstance(v, np.ndarray):
                batch_dict[k] = v.tolist()
            elif isinstance(v, torch.Tensor):
                batch_dict[k] = v.cpu().numpy().tolist()
        return batch_dict

    if dtype in ["np", "numpy"]:
        caster = np.asarray
        cast_type = np.ndarray
    elif dtype in ["pt", "torch", "pytorch"]:
        caster = torch.tensor
        cast_type = torch.Tensor
    else:
        raise ValueError(f"Invalid `dtype`: {dtype}")

    for k, v in batch_dict.items():
        if k not in skip_keys:
            try:
                if not isinstance(v, cast_type):
                    batch_dict[k] = caster(v)
            except Exception as e:  # noqa
                logger.warning(f"Could not convert values of `{k}` to type `{dtype}`\n" f"Error: {e}")
    return batch_dict


def get_non_numeric_keys(d: Dict, batched=True):
    """
    Get keys that have string values in a dictionary

    Args:
        d: The dict
        batched: Are the input dict values batched or not

    Returns:
        A list of string-valued keys
    """
    keys = []
    for k, v in d.items():
        if len(v) and isinstance(v[0], list):
            if batched and not isinstance(v[0][0], (int, float, complex)) and not isinstance(v[0][0], bool):
                keys.append(k)
            elif isinstance(v[0], str):
                keys.append(k)
    return keys



from dataclasses import dataclass, field
from typing import Iterable, List, Tuple

import numpy as np

from ..builders import build_preprocessor
from ..configs import PreprocessorConfig
from ..constants import (
    DEFAULT_IMAGE_PROCESSOR_CONFIG_FILE,
    DEFAULT_PREPROCESSOR_SUBFOLDER,
    Backends,
    ImageType,
)
from ..registry import register_preprocessor
from ..utils import (
    convert_batch_dict_dtype,
    convert_image_type,
    gray_scale_image,
    load_image,
    mirror_image,
    normalize_image,
    rescale_image,
    resize_image,
    transpose_channels_axis_side,
)
from .preprocessor import Preprocessor


_required_backends = [
    Backends.PILLOW,
]

_DESCRIPTION = r"""
A general image processor to do all the image transforms in a composable/configurable pipeline
"""

_image_type_aliases = {
    "pt": ImageType.TORCH,
    "pytorch": ImageType.TORCH,
    "torch": ImageType.TORCH,
    "np": ImageType.NUMPY,
    "numpy": ImageType.NUMPY,
    "pil": ImageType.PILLOW,
    "pillow": ImageType.PILLOW,
}


@dataclass
class ImageProcessorConfig(PreprocessorConfig):
    name = "image_processor"
    mean: List[float] = None
    std: List[float] = None
    rescale: float = None
    resample: int = None
    size: Tuple[int, int] = field(
        default=None,
        metadata={"description": "Image size tuple (width, height)"},
    )
    mirror: bool = False
    gray_scale: bool = False


@register_preprocessor("image_processor", config_class=ImageProcessorConfig, description=_DESCRIPTION)
class ImageProcessor(Preprocessor):
    """
    General image processor to perform sequential transforms on the images
    """

    required_backends = _required_backends

    preprocessor_subfolder = DEFAULT_PREPROCESSOR_SUBFOLDER
    image_processor_config_file = DEFAULT_IMAGE_PROCESSOR_CONFIG_FILE

    def __init__(self, config: ImageProcessorConfig, **kwargs):
        super().__init__(config, **kwargs)

    def __call__(
        self,
        images: List,
        device: str = None,
        mean: float = None,
        std: float = None,
        rescale: float = None,
        size: Tuple[int, int] = None,
        resample: float = None,
        mirror: bool = None,
        gray_scale: bool = None,
        return_tensors: str = "pt",
        **kwargs,
    ):
        """
        Perform a sequential image processing on a list of input images. You can control the behavior directly from
        this method's parameters.

        Args:
            images: A list of input images of types torch, numpy, pillow.
            mean: Image mean value for normalization
            std: Image std value for normalization
            rescale: Scale factor for rescaling the image
            size: A tuple of (width, height) to resize the images
            resample: Resample method value based on Image.Resampling
            return_tensors: The type of the output images
            **kwargs: Extra parameters

        Returns:
            The dict of transformed images list
        """
        mean = mean or self.config.mean
        std = std or self.config.std
        rescale = rescale or self.config.rescale
        size = size or self.config.size
        resample = resample or self.config.resample
        mirror = mirror or self.config.mirror
        gray_scale = gray_scale or self.config.gray_scale

        if isinstance(images, str):
            images = [images]
        # Load images if inputs are list of files
        images = [load_image(x, return_type="numpy") if isinstance(x, str) else x for x in images]

        # Cast image types
        images = [convert_image_type(image, target_type="numpy") for image in images]

        # Convert to grayscale
        if gray_scale:
            images = [gray_scale_image(image, return_type="numpy") for image in images]

        # Mirror images if mirror is set
        if mirror:
            images = [mirror_image(image, return_type="numpy") for image in images]

        if size is not None:
            if not isinstance(size, Iterable) or len(size) > 2:
                raise ValueError(f"The parameter `size` must be a tuple/list of (width, height), got `{size}`")
            images = [resize_image(image, size=size, resample=resample) for image in images]

        if rescale is not None:
            images = [rescale_image(image, scale=rescale) for image in images]

        if mean is not None and std is not None:
            images = [normalize_image(image, mean=mean, std=std, channel_axis="last") for image in images]

        # transpose channels axis
        images = [transpose_channels_axis_side(image, axis_side="first") for image in images]

        # Return images batch dict
        images = np.array([convert_image_type(image, target_type="numpy") for image in images], dtype=np.float32)

        images = convert_batch_dict_dtype({"pixel_values": images}, dtype=return_tensors)

        if device:
            import torch

            images = {k: v.to(device) if isinstance(v, torch.Tensor) else v for k, v in images.items()}

        return images

    @classmethod
    def load(
        cls,
        hub_or_local_path,
        subfolder: str = None,
        force_return_dict: bool = False,
        config_filename: str = None,
        **kwargs,
    ) -> "ImageProcessor":
        subfolder = subfolder or cls.preprocessor_subfolder
        config_filename = config_filename or cls.image_processor_config_file
        config = ImageProcessorConfig.load(hub_or_local_path, filename=config_filename, subfolder=subfolder)
        image_processor = build_preprocessor(config.name, config, **kwargs)
        return image_processor

    def save(
        self,
        path,
        subfolder=None,
        config_filename=None,
    ):
        subfolder = subfolder or self.preprocessor_subfolder
        config_filename = config_filename or self.image_processor_config_file

        self.config.save(path, subfolder=subfolder, filename=config_filename)

    def push_to_hub(
        self,
        repo_id,
        subfolder=None,
        commit_message=None,
        private=None,
        config_filename=None
    ):
        subfolder = subfolder or self.preprocessor_subfolder
        config_filename = config_filename or self.image_processor_config_file

        if commit_message is None:
            commit_message = "Hezar: Upload image processor files"

        self.config.push_to_hub(
            repo_id,
            subfolder=subfolder,
            filename=config_filename,
            private=private,
            commit_message=commit_message,
        )



from time import perf_counter


__all__ = [
    "exec_timer",
]


class exec_timer:
    """
    A context manager that captures the execution time of all the operations inside it

    Examples:
        >>> with exec_timer() as timer:
        >>>     # operations here
        >>> print(timer.time)
    """

    def __enter__(self):
        self.time = perf_counter()
        return self

    def __exit__(self, type, value, traceback):
        self.time = perf_counter() - self.time



import os
import tempfile
from typing import Dict, List, Union

from huggingface_hub import HfApi, hf_hub_download

from ..builders import build_embedding
from ..configs import EmbeddingConfig
from ..constants import (
    DEFAULT_EMBEDDING_CONFIG_FILE,
    DEFAULT_EMBEDDING_FILE,
    DEFAULT_EMBEDDING_SUBFOLDER,
    HEZAR_CACHE_DIR,
    Backends,
)
from ..utils import Logger, get_lib_version, verify_dependencies


logger = Logger(__name__)

# The below code is a workaround. Gensim's models have this limitation that the models can only be loaded using the same
# gensim & numpy version they were saved with.
REQUIRED_GENSIM_VERSION = "4.3.2"
REQUIRED_NUMPY_VERSION = "1.24"


# Check if the right combo of gensim/numpy versions are installed
def _verify_gensim_installation():
    if (
        not get_lib_version("numpy").startswith(REQUIRED_NUMPY_VERSION)
        or not get_lib_version("gensim").startswith(REQUIRED_GENSIM_VERSION)
    ):
        raise ImportError(
            f"The required Gensim version for this version of Hezar is currently {REQUIRED_GENSIM_VERSION} "
            f"and the required Numpy version for Gensim=={REQUIRED_GENSIM_VERSION} is {REQUIRED_NUMPY_VERSION}"
        )


class Embedding:
    """
    Base class for all embeddings.
    """

    required_backends: List[Union[str, Backends]] = []

    filename = DEFAULT_EMBEDDING_FILE
    vectors_filename = f"{filename}.wv.vectors.npy"
    config_filename = DEFAULT_EMBEDDING_CONFIG_FILE
    subfolder = DEFAULT_EMBEDDING_SUBFOLDER

    def __init__(self, config: EmbeddingConfig, embedding_file: str = None, vectors_file: str = None, **kwargs):
        verify_dependencies(self, self.required_backends)  # Check if all the required dependencies are installed
        _verify_gensim_installation()

        self.config = config.update(kwargs)
        self.model = self.from_file(embedding_file, vectors_file) if embedding_file else self.build()

    def build(self):
        raise NotImplementedError

    def from_file(self, embedding_path, vectors_path):
        raise NotImplementedError

    def __call__(self, inputs: Union[str, List[str]], **kwargs):
        if isinstance(inputs, str):
            inputs = [inputs]
        vectors = [self.word_vectors[w] for w in inputs]
        return vectors

    def train(
        self,
        dataset,
        epochs,
    ):
        raise NotImplementedError

    def word2index(self, word):
        return self.vocab.get(word, -1)

    def index2word(self, index):
        keyed_vocab = {v: k for k, v in self.vocab.items()}
        return keyed_vocab[index]

    def similarity(self, word1: str, word2: str):
        raise NotImplementedError

    def doesnt_match(self, words: List[str]):
        raise NotImplementedError

    def most_similar(self, word: str, top_n: int = 5):
        raise NotImplementedError

    def get_normed_vectors(self):
        raise NotImplementedError

    @classmethod
    def load(
        cls,
        hub_or_local_path,
        config_filename=None,
        embedding_file=None,
        vectors_file=None,
        subfolder=None,
        **kwargs,
    ) -> "Embedding":
        config_filename = config_filename or cls.config_filename
        embedding_file = embedding_file or cls.filename
        vectors_file = vectors_file or cls.vectors_filename
        subfolder = subfolder or cls.subfolder

        config = EmbeddingConfig.load(hub_or_local_path, filename=config_filename, subfolder=subfolder)

        if os.path.isdir(hub_or_local_path):
            embedding_path = os.path.join(hub_or_local_path, subfolder, embedding_file)
            vectors_path = os.path.join(hub_or_local_path, subfolder, vectors_file)
        else:
            embedding_path = hf_hub_download(
                hub_or_local_path,
                filename=embedding_file,
                subfolder=subfolder,
                cache_dir=HEZAR_CACHE_DIR,
                resume_download=True,
            )
            vectors_path = hf_hub_download(
                hub_or_local_path,
                filename=vectors_file,
                subfolder=subfolder,
                cache_dir=HEZAR_CACHE_DIR,
                resume_download=True,
            )

        embedding = build_embedding(
            config.name,
            config=config,
            embedding_file=embedding_path,
            vectors_file=vectors_path,
            **kwargs,
        )

        return embedding

    def save(
        self,
        path: Union[str, os.PathLike],
        filename: str = None,
        subfolder: str = None,
        save_config: bool = True,
        config_filename: str = None,
    ):
        raise NotImplementedError

    def push_to_hub(
        self,
        repo_id,
        commit_message=None,
        subfolder=None,
        filename=None,
        vectors_filename=None,
        config_filename=None,
        private=False,
    ):
        subfolder = subfolder or self.subfolder
        filename = filename or self.filename
        vectors_filename = vectors_filename or self.vectors_filename
        config_filename = config_filename or self.config_filename

        api = HfApi()
        # create remote repo
        api.create_repo(repo_id, exist_ok=True)
        # save to tmp and prepare for push
        cache_path = tempfile.mkdtemp()
        # save embedding model file
        embedding_save_dir = os.path.join(cache_path)
        os.makedirs(embedding_save_dir, exist_ok=True)

        if commit_message is None:
            commit_message = "Hezar: Upload embedding and config"

        self.save(embedding_save_dir, filename, subfolder=subfolder, save_config=False)

        self.config.push_to_hub(
            repo_id,
            config_filename,
            subfolder=subfolder,
            repo_type="model",
            private=private,
            commit_message=commit_message,
        )

        api.upload_file(
            repo_id=repo_id,
            path_or_fileobj=os.path.join(embedding_save_dir, subfolder, filename),
            repo_type="model",
            path_in_repo=f"{subfolder}/{filename}",
            commit_message=commit_message,
        )
        logger.log_upload_success(
            name=f"{self.__class__.__name__}(name={self.config.name})",
            target_path=f"{os.path.join(repo_id, subfolder, filename)}",
        )

        api.upload_file(
            repo_id=repo_id,
            path_or_fileobj=os.path.join(embedding_save_dir, subfolder, vectors_filename),
            repo_type="model",
            path_in_repo=f"{subfolder}/{vectors_filename}",
            commit_message=commit_message,
        )
        logger.log_upload_success(
            name=f"`{self.__class__.__name__}(name={self.config.name})`",
            target_path=f"`{os.path.join(repo_id, subfolder, vectors_filename)}`",
        )

    def torch_embedding(self):
        import torch

        weights = torch.FloatTensor(self.vectors)
        embedding_layer = torch.nn.Embedding.from_pretrained(weights)
        return embedding_layer

    @property
    def word_vectors(self):
        """
        Get key:value pairs of word:vector
        """
        raise NotImplementedError

    @property
    def vectors(self):
        """
        Get the all vectors array/tensor
        """
        raise NotImplementedError

    @property
    def vocab(self) -> Dict[str, int]:
        raise NotImplementedError



from dataclasses import dataclass

from datasets import load_dataset

from ...configs import DatasetConfig
from ...constants import TaskType
from ...preprocessors import Tokenizer
from ...registry import register_dataset
from ...utils import Logger
from ..data_collators import TextPaddingDataCollator
from .dataset import Dataset


logger = Logger(__name__)


@dataclass
class TextSummarizationDatasetConfig(DatasetConfig):
    name = "text_summarization"
    task: str = TaskType.TEXT_GENERATION
    path: str = None
    tokenizer_path: str = None
    prefix: str = None
    text_field: str = None
    summary_field: str = None
    title_field: str = None
    max_length: int = None


@register_dataset("text_summarization", config_class=TextSummarizationDatasetConfig)
class TextSummarizationDataset(Dataset):
    """
    A text summarization dataset class.
    As of now this class is intended for datasets existing on the Hub!

    Args:
        config: Dataset config obj
        split: Which split to use
        **kwargs: Extra config parameters to assign to the original config
    """

    def __init__(self, config: TextSummarizationDatasetConfig, split=None, **kwargs):
        super().__init__(config, **kwargs)
        self.dataset = self._load(split)
        self.tokenizer = self._build_tokenizer()
        self.data_collator = TextPaddingDataCollator(
            tokenizer=self.tokenizer,
            max_length=self.config.max_length,
            padding_type="max_length" if self.config.max_length else "longest",
        )

    def _load(self, split):
        """
        Load the dataset

        Args:
            split: Dataset split

        Returns:
            The whole dataset
        """
        # TODO: In case we want to make this class work on other types like csv, json, etc. we have to do it here.
        dataset = load_dataset(self.config.path, split=split, cache_dir=self.cache_dir)
        return dataset

    def _build_tokenizer(self):
        if self.config.tokenizer_path:
            tokenizer = Tokenizer.load(self.config.tokenizer_path)
        else:
            logger.warning(
                "This dataset requires a tokenizer to work. Provide it in config as `tokenizer_path` "
                "or set it manually as `dataset.tokenizer = your_tokenizer` after building the dataset."
            )
            tokenizer = None
        return tokenizer

    def __len__(self):
        return len(self.dataset)

    def __getitem__(self, index):
        """
        Tokenize inputs and return a dict containing ids, masks, labels, etc.

        Args:
            index: Sample index

        Returns:
            A dict of tokenized text data and labels and some extra stuff
        """
        text = self.dataset[index][self.config.text_field]
        if self.config.prefix is not None:
            text = self.config.prefix + text  # for conditional generation we might need a static prefix
        summary = self.dataset[index][self.config.summary_field]

        inputs = self.tokenizer(
            text,
            return_tensors="pt",
            max_length=self.config.max_length,
            padding="max_length" if self.config.max_length else "longest",
            return_attention_mask=True,
        )
        labels = self.tokenizer(
            summary,
            return_tensors="pt",
            max_length=self.config.max_length,
            padding="max_length" if self.config.max_length else "longest",
            return_attention_mask=True,
        )

        inputs["labels"] = labels["token_ids"]

        return inputs



import json
import os
from typing import Dict, Union

import omegaconf
from omegaconf import DictConfig

from .logging import Logger


__all__ = [
    "flatten_dict",
    "load_yaml_config",
    "load_json_config",
    "hezar_config_to_hf_config",
]

logger = Logger(__name__)


def flatten_dict(dict_config: Union[Dict, DictConfig]) -> DictConfig:
    """
    Flatten a nested Dict/DictConfig object

    Args:
        dict_config: A Dict/DictConfig object

    Returns:
        The flattened version of the dict-like object
    """

    config = DictConfig({})
    for k, v in dict_config.items():
        if isinstance(v, (Dict, DictConfig)):
            config.update(flatten_dict(v))
        else:
            config[k] = v

    return config


def load_yaml_config(path: Union[str, os.PathLike]):
    """
    Load yaml file using omegaconf
    """
    config = omegaconf.OmegaConf.load(path)
    return config


def load_json_config(path: Union[str, os.PathLike]):
    """
    Load json config file
    """
    with open(path) as f:
        config = json.load(f)
    return config


def hezar_config_to_hf_config(config):
    """
    Convert a :class:`hezar.Config` instance to HuggingFace :class:`transformers.PretrainedConfig` format

    Args:
        config: A :class:`Hezar.Config` instance
    """
    from transformers import PretrainedConfig

    hf_config = PretrainedConfig(**config)
    return hf_config



import gzip
import shutil

from .logging import Logger


logger = Logger(__name__)

__all__ = ["gunzip"]


def gunzip(src_path, dest_path):
    """
    Unzip a .gz file from `src_path` and extract to `dest_path`
    Args:
        src_path: Path to .gz file
        dest_path: Path to the destination file

    Returns:

    """
    with gzip.open(src_path, "rb") as f_in:
        with open(dest_path, "wb") as f_out:
            shutil.copyfileobj(f_in, f_out)
    logger.debug(f"Extracted {src_path} to {dest_path}")



from typing import Iterable, Tuple, Union

import numpy as np
import torch

from ..constants import Backends, ChannelsAxisSide, ImageType
from .integration_utils import is_backend_available
from .logging import Logger


logger = Logger(__name__)

if is_backend_available(Backends.PILLOW):
    from PIL import Image

__all__ = [
    "convert_image_type",
    "normalize_image",
    "load_image",
    "show_image",
    "rescale_image",
    "resize_image",
    "mirror_image",
    "gray_scale_image",
    "find_channels_axis_side",
    "transpose_channels_axis_side",
]


def verify_image_dims(image: np.ndarray):
    if len(image.shape) not in (2, 3):
        raise ValueError(f"Image input must be a numpy array of size 2 or 3! Got {image.shape}")


def convert_image_type(
    image: Union[np.ndarray, "Image", torch.Tensor],
    target_type: Union[str, ImageType] = ImageType.NUMPY,
):
    """
    Convert image lib type. Supports numpy array, pillow image and torch tensor.
    """
    if isinstance(image, Image.Image):
        if image.mode == "L":
            image = np.asarray(image)
            image = np.expand_dims(image, 0)
        else:
            image = np.asarray(image)
    elif isinstance(image, torch.Tensor):
        image = image.cpu().numpy()

    verify_image_dims(image)

    if target_type == ImageType.PILLOW:
        # transpose channels to the last axis since pillow cannot handle it otherwise
        if find_channels_axis_side(image) == ChannelsAxisSide.FIRST:
            image = transpose_channels_axis_side(
                image,
                axis_side=ChannelsAxisSide.LAST,
                src_axis_side=ChannelsAxisSide.FIRST,
            )
        num_channels = image.shape[0] if find_channels_axis_side(image) == ChannelsAxisSide.FIRST else image.shape[-1]
        if num_channels == 1:
            image = image[:, :, -1]
            image = Image.fromarray(image, "L")
        else:
            image = Image.fromarray(image)
    elif target_type == ImageType.TORCH:
        image = torch.tensor(image)

    return image


def load_image(path, return_type: Union[str, ImageType] = ImageType.PILLOW):
    """
    Load an image file to a desired return format

    Args:
        path: Path to image file
        return_type: Image output type ("pillow", "numpy", "torch")

    Returns:
        The desired output image of type `PIL.Image` or `numpy.ndarray` or `torch.Tensor`
    """
    pil_image = Image.open(path).convert("RGB")
    converted_image = convert_image_type(pil_image, return_type)
    return converted_image


def show_image(image: Union["Image", torch.Tensor, np.ndarray], title: str = "Image"):
    """
    Given any type of input image (PIL, numpy, torch), show the image in a window

    Args:
        image: Input image of types PIL.Image, numpy.ndarray or torch.Tensor
        title: Optional title for the preview window
    """
    pil_image = convert_image_type(image, ImageType.PILLOW)
    pil_image.show(title=title)


def rescale_image(image: np.ndarray, scale: float):
    verify_image_dims(image)
    image = image * scale
    return image


def resize_image(
    image: np.ndarray,
    size: Tuple[int, int],
    resample=None,
    reducing_gap: float = None,
    return_type: ImageType = ImageType.NUMPY,
):
    """
    Resize a numpy array image (actually uses pillow PIL.Image.resize(...))

    Args:
        image: Numpy image
        size: A tuple of (width, height)
        resample: Resampling filter (refer to PIL.Image.Resampling) for possible values
        reducing_gap: Optimization method for resizing based on reducing times
        return_type: Return type of the image (numpy, torch, pillow)

    Returns:
        The resized image
    """
    verify_image_dims(image)
    if len(size) != 2:
        raise ValueError(f"The value of `size` must be a 2-sized tuple! Got length {len(size)}(`{size}`)")
    pil_image = convert_image_type(image, ImageType.PILLOW)
    pil_image = pil_image.resize(size, resample=resample, reducing_gap=reducing_gap)
    np_image = convert_image_type(pil_image, return_type)
    return np_image


def mirror_image(image: np.ndarray, return_type: Union[str, ImageType] = ImageType.NUMPY):
    if not isinstance(image, np.ndarray):
        raise ValueError("image must be a numpy array")

    verify_image_dims(image)

    pil_image = convert_image_type(image, ImageType.PILLOW)
    pil_image = pil_image.transpose(Image.FLIP_LEFT_RIGHT)
    final_image = convert_image_type(pil_image, return_type)
    return final_image


def gray_scale_image(image: np.ndarray, return_type: Union[str, ImageType] = ImageType.NUMPY):
    if not isinstance(image, np.ndarray):
        raise ValueError("image must be a numpy array")

    verify_image_dims(image)

    pil_image = convert_image_type(image, ImageType.PILLOW)
    pil_image = pil_image.convert("L")
    np_image = convert_image_type(pil_image, ImageType.NUMPY)
    final_image = convert_image_type(np_image, target_type=return_type)
    return final_image


def normalize_image(
    image: np.ndarray,
    mean: Union[float, Iterable[float]],
    std: Union[float, Iterable[float]],
    channel_axis: Union[str, ChannelsAxisSide] = "first",
):
    verify_image_dims(image)

    if not isinstance(image, np.ndarray):
        raise ValueError("image must be a numpy array")

    num_channels = image.shape[0 if channel_axis == ChannelsAxisSide.FIRST else -1]

    if not isinstance(mean, Iterable):
        mean = [mean] * num_channels
    mean = np.array(mean, dtype=image.dtype)

    if not isinstance(std, Iterable):
        std = [std] * num_channels
    std = np.array(std, dtype=image.dtype)

    if channel_axis == ChannelsAxisSide.LAST:
        image = (image - mean) / std
    else:
        image = ((image.T - mean) / std).T

    return image


def find_channels_axis_side(image: np.ndarray, num_channels: int = None):
    valid_num_channels = (num_channels,) if num_channels is not None else (1, 2, 3)
    if image.shape[0] in valid_num_channels:
        return ChannelsAxisSide.FIRST
    else:
        return ChannelsAxisSide.LAST


def transpose_channels_axis_side(
    image: np.ndarray,
    axis_side: Union[str, ChannelsAxisSide],
    num_channels: int = None,
    src_axis_side: Union[str, ChannelsAxisSide] = None,
):
    """
    Convert an image channels axis side from (channels, ...) to (..., channels) or vise versa.

    Args:
        image: Input image
        axis_side: The desired axis side (can be "first" or "last")
        num_channels: The number of channels in the input image
        src_axis_side: The image initial channels axis side (can be "first" or "last")

    Returns:
        The image with the converted channels axis
    """
    if src_axis_side is None:
        src_axis_side = find_channels_axis_side(image, num_channels=num_channels)

    # If input's channels axis side and output channels axis side are the same return the same image
    if src_axis_side == axis_side:
        return image

    if axis_side == ChannelsAxisSide.FIRST:
        image = image.transpose((2, 0, 1))
    elif axis_side == ChannelsAxisSide.LAST:
        image = image.transpose((1, 2, 0))

    return image



from dataclasses import dataclass, field
from typing import List

from ....configs import ModelConfig


@dataclass
class WhisperSpeechRecognitionConfig(ModelConfig):
    name = "whisper_speech_recognition"
    vocab_size: int = 51865
    num_mel_bins: int = 80
    encoder_layers: int = 6
    encoder_attention_heads: int = 4
    decoder_layers: int = 6
    decoder_attention_heads: int = 4
    num_hidden_layers: int = 12
    decoder_ffn_dim: int = 1536
    encoder_ffn_dim: int = 1536
    encoder_layerdrop: float = 0.0
    decoder_layerdrop: float = 0.0
    decoder_start_token_id: int = 50257
    use_cache: bool = True
    sampling_rate: int = 16000
    is_encoder_decoder: bool = True
    activation_function: str = "gelu"
    d_model: int = 256
    dropout: float = 0.0
    torch_dtype: str = "float32"
    attention_dropout: float = 0.0
    activation_dropout: float = 0.0
    init_std: float = 0.02
    scale_embedding: bool = False
    max_source_positions: int = 1500
    max_target_positions: int = 448
    pad_token_id: int = 50256
    bos_token_id: int = 50257
    eos_token_id: int = 50256
    suppress_tokens: List[int] = None
    begin_suppress_tokens: List[int] = field(default_factory=lambda: [220, 50256])
    use_weighted_layer_sum: bool = False
    classifier_proj_size: int = 256
    apply_spec_augment: bool = False
    mask_time_prob: float = 0.05
    mask_time_length: int = 10
    mask_time_min_masks: int = 2
    mask_feature_prob: float = 0.0
    mask_feature_length: int = 10
    mask_feature_min_masks: int = 0
    max_new_tokens: int = 448



from typing import Dict, List, Union

from ..configs import MetricConfig
from ..constants import Backends
from ..utils import verify_dependencies


class Metric:
    """
    The base metric class for all metrics in Hezar.

    Metrics are simple wrappers for casual ready-to-use metrics like in scikit-learn, etc. and it's strongly recommended
    not to reinvent the wheel. If a metric is already implemented by some package, use it! The only reason to implement
    such a module, is to make sure the metrics are treated the same all over the framework.
    """

    required_backends: Union[List[Union[str, Backends]]] = []

    def __init__(self, config: MetricConfig, **kwargs):
        # Check if all the required dependencies are installed
        verify_dependencies(self, self.required_backends)

        self.config = config.update(kwargs)

    def compute(self, predictions=None, targets=None, **kwargs) -> Dict:
        """
        Compute metric value for the given predictions against the targets
        Args:
            predictions: Prediction values
            targets: Ground truth values
            **kwargs: Extra arguments depending on the metric

        Returns:
            A dictionary of the results and scores
        """
        raise NotImplementedError



from .registry import *
from .builders import *
from .configs import *
from .data import *
from .embeddings import *
from .metrics import *
from .models import *
from .preprocessors import *
from .trainer import *
from .utils import *


__version__ = "0.32.1"



from dataclasses import dataclass

from ....configs import ModelConfig
from ....constants import TaskType


@dataclass
class DistilBERTConfig(ModelConfig):
    name = "distilbert"
    task: str = TaskType.LANGUAGE_MODELING
    activation: str = "gelu"
    attention_dropout: float = 0.1
    dim: int = 768
    dropout: float = 0.1
    hidden_dim: int = 3072
    initializer_range: float = 0.02
    max_position_embeddings: int = 512
    n_heads: int = 12
    n_layers: int = 6
    output_past: bool = True
    pad_token_id: int = 0
    qa_dropout: float = 0.1
    tie_weights_: bool = True
    vocab_size: int = 42000



from dataclasses import dataclass

from ..configs import MetricConfig
from ..constants import Backends, MetricType
from ..registry import register_metric
from ..utils import is_backend_available
from .metric import Metric


if is_backend_available(Backends.JIWER):
    import jiwer

_DESCRIPTION = "Word Error Rate (WER) using `jiwer`. Commonly used for Speech Recognition systems"

_required_backends = [
    Backends.JIWER,
]


@dataclass
class WERConfig(MetricConfig):
    name = MetricType.WER
    concatenate_texts: bool = False
    output_keys: tuple = ("wer",)


@register_metric("wer", config_class=WERConfig, description=_DESCRIPTION)
class WER(Metric):
    required_backends = _required_backends

    def __init__(self, config: WERConfig, **kwargs):
        super().__init__(config=config, **kwargs)

    def compute(
        self,
        predictions=None,
        targets=None,
        concatenate_texts=None,
        n_decimals=None,
        output_keys=None,
        **kwargs,
    ):
        concatenate_texts = concatenate_texts or self.config.concatenate_texts
        n_decimals = n_decimals or self.config.n_decimals

        if concatenate_texts:
            score = jiwer.compute_measures(targets, predictions)["wer"]
        else:
            incorrect = 0
            total = 0
            for prediction, reference in zip(predictions, targets):
                measures = jiwer.compute_measures(reference, prediction)
                incorrect += measures["substitutions"] + measures["deletions"] + measures["insertions"]
                total += measures["substitutions"] + measures["deletions"] + measures["hits"]

            score = incorrect / total

        results = {"wer": round(float(score), n_decimals)}

        if output_keys:
            results = {k: v for k, v in results.items() if k in output_keys}

        return results



from dataclasses import dataclass
from typing import List, Optional, Union

from ..configs import MetricConfig
from ..constants import Backends, MetricType
from ..registry import register_metric
from ..utils import Logger, is_backend_available
from .metric import Metric


if is_backend_available(Backends.SEQEVAL):
    from seqeval.metrics import accuracy_score, classification_report

_required_backends = [
    Backends.SEQEVAL,
]

logger = Logger(__name__)


@dataclass
class SeqevalConfig(MetricConfig):
    name = MetricType.SEQEVAL
    output_keys: tuple = ("accuracy", "recall", "precision", "f1")
    suffix: bool = False
    mode: Optional[str] = None
    sample_weight: Optional[List[int]] = None
    zero_division: Union[str, int] = 0


@register_metric("seqeval", config_class=SeqevalConfig)
class Seqeval(Metric):
    required_backends = _required_backends

    def __init__(self, config: SeqevalConfig, **kwargs):
        super().__init__(config, **kwargs)

    def compute(
        self,
        predictions=None,
        targets=None,
        suffix: bool = None,
        mode: Optional[str] = None,
        sample_weight: Optional[List[int]] = None,
        zero_division: Union[str, int] = None,
        n_decimals: int = None,
        output_keys=None,
        **kwargs,
    ):
        suffix = suffix or self.config.suffix
        mode = mode or self.config.mode
        sample_weight = sample_weight or self.config.sample_weight
        zero_division = zero_division or self.config.zero_division
        n_decimals = n_decimals or self.config.n_decimals
        output_keys = output_keys or self.config.output_keys

        report = classification_report(
            y_true=targets,
            y_pred=predictions,
            suffix=suffix,
            output_dict=True,
            mode=mode,
            sample_weight=sample_weight,
            zero_division=zero_division,
        )
        report.pop("macro avg")
        report.pop("weighted avg")
        overall_score = report.pop("micro avg")

        results = {
            "accuracy": format(accuracy_score(predictions, targets)),
            "f1": overall_score["f1-score"],
            "recall": overall_score["recall"],
            "precision": overall_score["precision"],
        }

        results = {k: round(float(v), n_decimals) for k, v in results.items()}

        if output_keys:
            results = {k: v for k, v in results.items() if k in output_keys}

        return results



import os
from typing import List, Optional, Union

from torch.utils.data import Dataset as TorchDataset

from ...builders import build_dataset
from ...configs import DatasetConfig
from ...constants import (
    DEFAULT_DATASET_CONFIG_FILE,
    HEZAR_CACHE_DIR,
    Backends,
    RepoType,
    SplitType,
)
from ...utils import verify_dependencies


class Dataset(TorchDataset):
    """
    Base class for all datasets in Hezar.

    Args:
        config:
        **kwargs:
    """
    required_backends: List[Union[str, Backends]] = None
    config_filename = DEFAULT_DATASET_CONFIG_FILE
    cache_dir = HEZAR_CACHE_DIR

    def __init__(self, config: DatasetConfig, **kwargs):
        verify_dependencies(self, self.required_backends)
        self.config = config.update(kwargs)
        self.preprocessor = None
        self.data_collator = None

    def __len__(self):
        raise NotImplementedError

    def __getitem__(self, index):
        raise NotImplementedError

    @classmethod
    def load(
        cls,
        hub_path: Union[str, os.PathLike],
        config_filename: Optional[str] = None,
        split: Optional[Union[str, SplitType]] = None,
        **kwargs,
    ) -> "Dataset":
        """
        Load the dataset from hub.

        Args:
            hub_path: Path to dataset from hub or locally
            config_filename: Dataset config file name
            split: Dataset split, defaults to "train"
            **kwargs: Config parameters as keyword arguments

        Returns:

        """
        split = split or "train"
        config_filename = config_filename or cls.config_filename
        dataset_config = DatasetConfig.load(hub_path, filename=config_filename, repo_type=RepoType.DATASET)
        dataset_config.path = hub_path
        dataset = build_dataset(dataset_config.name, config=dataset_config, split=split, **kwargs)
        return dataset



import os
from collections import OrderedDict
from typing import List, Union

from huggingface_hub import hf_hub_download
from omegaconf import OmegaConf

from ..constants import DEFAULT_PREPROCESSOR_SUBFOLDER, Backends, RegistryType, RepoType
from ..utils import get_module_class, list_repo_files, verify_dependencies


class Preprocessor:
    """
    Base class for all data preprocessors.

    Args:
        config: Preprocessor properties
    """

    required_backends: List[Union[str, Backends]] = []

    preprocessor_subfolder = DEFAULT_PREPROCESSOR_SUBFOLDER

    def __init__(self, config, **kwargs):
        verify_dependencies(self, self.required_backends)  # Check if all the required dependencies are installed

        self.config = config.update(kwargs)

    def __call__(self, inputs, **kwargs):
        """
        An abstract call method for a preprocessor. All preprocessors must implement this.

        Args:
            inputs: Raw inputs to process. Usually a list or a dict
            **kwargs: Extra keyword arguments depending on the preprocessor
        """
        raise NotImplementedError

    def save(self, path, **kwargs):
        raise NotImplementedError

    def push_to_hub(
        self,
        repo_id,
        subfolder=None,
        commit_message=None,
        private=None,
        **kwargs,
    ):
        raise NotImplementedError

    @classmethod
    def load(
        cls,
        hub_or_local_path,
        subfolder: str = None,
        force_return_dict: bool = False,
        **kwargs
    ):
        """
        Load a preprocessor or a pipeline of preprocessors from a local or Hub path. This method automatically detects
        any preprocessor in the path. If there's only one preprocessor, returns it and if there are more, returns a
        dictionary of preprocessors.

        This method must also be overriden by subclasses as it internally calls this method for every possible
        preprocessor found in the repo.

        Args:
            hub_or_local_path: Path to hub or local repo
            subfolder: Subfolder for the preprocessor.
            force_return_dict: Whether to return a dict even if there's only one preprocessor available on the repo
            **kwargs: Extra kwargs

        Returns:
            A Preprocessor subclass or a dict of Preprocessor subclass instances
        """
        subfolder = subfolder or cls.preprocessor_subfolder
        preprocessor_files = list_repo_files(hub_or_local_path, subfolder=subfolder)
        preprocessors = PreprocessorsContainer()
        for f in preprocessor_files:
            if f.endswith(".yaml"):
                if os.path.isdir(hub_or_local_path):
                    config_file = os.path.join(hub_or_local_path, subfolder, f)
                else:
                    config_file = hf_hub_download(
                        hub_or_local_path,
                        filename=f,
                        subfolder=subfolder,
                        repo_type=RepoType.MODEL
                    )
                config = OmegaConf.load(config_file)
                name = config.get("name", None)
                if name:
                    preprocessor_cls = get_module_class(name, registry_type=RegistryType.PREPROCESSOR)
                    preprocessor = preprocessor_cls.load(hub_or_local_path, subfolder=subfolder)
                    preprocessors[name] = preprocessor
                else:
                    raise ValueError(f"The config file `{config_file}` does not have the property `name`!")
        if len(preprocessors) == 1 and not force_return_dict:
            return list(preprocessors.values())[0]

        return preprocessors


class PreprocessorsContainer(OrderedDict):
    """
    A class to hold the preprocessors by their name
    """

    def __getattr__(self, item):
        """
        Override this method to be able to access preprocessors as attributes

        Examples:
            >>> preprocessor["text_normalizer"] is preprocessor.text_normalizer  # noqa
            ... True
        """
        if item in self:
            return self[item]
        else:
            super().__getattribute__(item)

    def save(self, path):
        """
        Save every preprocessor item in the container
        """
        for name, preprocessor in self.items():
            preprocessor.save(path)

    def push_to_hub(
        self,
        repo_id,
        subfolder=None,
        commit_message=None,
        private=None,
    ):
        """
        Push every preprocessor item in the container
        """
        for name, preprocessor in self.items():
            preprocessor.push_to_hub(repo_id, subfolder=subfolder, commit_message=commit_message, private=private)



from dataclasses import dataclass

from ....configs import ModelConfig
from ....constants import TaskType


@dataclass
class RobertaTextClassificationConfig(ModelConfig):
    name = "roberta_text_classification"
    task: str = TaskType.TEXT_CLASSIFICATION
    num_labels: int = None
    id2label: dict = None
    attention_probs_dropout_prob: float = 0.1
    bos_token_id: int = 0
    eos_token_id: int = 2
    gradient_checkpointing: bool = False
    hidden_act: str = "gelu"
    hidden_dropout_prob: float = 0.1
    hidden_size: int = 768
    classifier_dropout: float = None
    initializer_range: int = 0.02
    intermediate_size: int = 3072
    layer_norm_eps: float = 1e-12
    max_position_embeddings: int = 514
    num_attention_heads: int = 12
    num_hidden_layers: int = 12
    pad_token_id: int = 1
    position_embedding_type: str = "absolute"
    type_vocab_size: int = 1
    use_cache: bool = True
    vocab_size: int = 42000
